# -*- coding: utf-8 -*-


"""
The general functionalities that all querybuilders need to have
are found in this module.
:func:`AbstractQueryBuilder` is the abstract class for QueryBuilder classes.
Subclasses need to be written for *every* schema/backend implemented
in backends.

"""

import copy
import datetime
import warnings
from abc import abstractmethod, ABCMeta
from inspect import isclass as inspect_isclass
from sa_init import (
        aliased, and_, or_, not_, func as sa_func,
        InstrumentedAttribute, Cast
    )
from aiida.common.exceptions import (
        InputValidationError, DbContentError, MissingPluginError
    )
from aiida.common.utils import flatten_list
from aiida.common.hashing import make_hash



__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

class AbstractQueryBuilder(object):
    """
    QueryBuilderBase is the base class for QueryBuilder classes,
    which are than adapted to the individual schema and ORM used.
    In here, general graph traversal functionalities are implemented,
    the specific type of node and link is dealt in subclasses.
    In order to load the correct subclass::

        from aiida.orm.querybuilder import QueryBuilder
    """

    __metaclass__ = ABCMeta


    _EDGE_TAG_DELIM = '--'
    _VALID_PROJECTION_KEYS = ('func', 'cast')


    def __init__(self, *args, **kwargs):

        # A list storing the path being traversed by the query
        self._path = []

        # The list of unique tags
        # self.tag_list = []# not needed any more

        # A list of unique aliases in same order as path
        self._aliased_path = []

        # A dictionary tag:alias of ormclass
        # redundant but makes life easier
        self._tag_to_alias_map = {}

        # A dictionary tag: filter specification for this alias
        self._filters = {}

        # A dictionary tag: projections for this alias
        self._projections = {}

        # A dictionary for classes passed to the tag given to them
        # Everything is specified with unique tags, which are strings.
        # But somebody might not care about giving tags, so to do
        # everything with classes one needs a map
        # qb = QueryBuilder(path = [PwCalculation])
        # qb.append(StructureData, input_of=PwCalculation

        # The cls_to_tag_map in this case would be:
        # {PwCalculation:'PwCalculation', StructureData:'StructureData'}

        self._cls_to_tag_map = {}

        self._hash = None

        self._injected = False
        if args:
            raise InputValidationError(
                    "Arguments are not accepted\n"
                    "when instantiating a QueryBuilder instance"
                )

        path = kwargs.pop('path', [])
        if not isinstance(path, (tuple, list)):
            raise InputValidationError(
                    "Path needs to be a tuple or a list"
                )
        for path_spec in path:
            try:
                self.append(**path_spec)
            except TypeError as e:
                if isinstance(path_spec, basestring):
                    # Maybe it is just a string,
                    # I assume user means the type
                    self.append(type=path_spec)
                else:
                    # Or a class, let's try
                    self.append(cls=path_spec)

        for key, val in kwargs.pop('project', {}).items():
            self.add_projection(key, val)

        for key, val in kwargs.pop('filters', {}).items():
            self.add_filter(key, val)

        self.limit(kwargs.pop('limit', None))

        self.offset(kwargs.pop('offset', None))

        self._order_by = {}
        order_spec = kwargs.pop('order_by', None)
        if order_spec:
            self.order_by(order_spec)

        if kwargs:
            valid_keys = [
                    'path', 'filters', 'project', 'limit', 'order_by'
            ]
            raise InputValidationError(
                    "Received additional keywords: {}"
                    "\nwhich I cannot process"
                    "\nValid keywords are: {}"
                    "".format(kwargs.keys(), valid_keys)
            )


    def __str__(self):
        from aiida.common.setup import get_profile_config
        from aiida.backends import settings
        from aiida.common.exceptions import ConfigurationError

        engine = get_profile_config(settings.AIIDADB_PROFILE)["AIIDADB_ENGINE"]

        if engine == "sqlite3":
            from sqlalchemy.dialects import sqlite as mydialect
        elif engine.startswith("mysql"):
            from sqlalchemy.dialects import mysql as mydialect
        elif engine.startswith("postgre"):
            from sqlalchemy.dialects import postgresql as mydialect
        else:
            raise ConfigurationError("Unknown DB engine: {}".format(
                    engine))

        que = self.get_query()
        return str(que.statement.compile(
                compile_kwargs={"literal_binds": True},
                dialect=mydialect.dialect()
            )
        )



    def _get_ormclass(self, cls, ormclasstype):
        """
        Return the valid ormclass for the connections
        """
        # Checks whether valid cls and ormclasstype are done before

        # If it is a class:
        if cls:
            # Nodes:
            if issubclass(cls, self.Node):
                # If something pass an ormclass node
                # Users wouldn't do that, by why not...
                ormclasstype = self.AiidaNode._plugin_type_string
                query_type_string = self.AiidaNode._query_type_string
                ormclass = cls
            elif issubclass(cls, self.AiidaNode):
                ormclasstype = cls._plugin_type_string
                query_type_string = cls._query_type_string
                ormclass = self.Node
            # Groups:
            elif issubclass(cls, self.Group):
                ormclasstype = 'group'
                query_type_string = None
                ormclass = cls
            elif issubclass(cls, self.AiidaGroup):
                ormclasstype = 'group'
                query_type_string = None
                ormclass = self.Group
            # Computers:
            elif issubclass(cls, self.Computer):
                ormclasstype = 'computer'
                query_type_string = None
                ormclass = cls
            elif issubclass(cls, self.AiidaComputer):
                ormclasstype = 'computer'
                query_type_string = None
                ormclass = self.Computer

            # Users
            elif issubclass(cls, self.User):
                ormclasstype = 'user'
                query_type_string = None
                ormclass = cls
            elif issubclass(cls, self.AiidaUser):
                ormclasstype = 'user'
                query_type_string = None
                ormclass = self.User
            else:
                raise InputValidationError(
                        "\n\n\n"
                        "I do not know what to do with {}"
                        "\n\n\n".format(cls)
                    )
        # If it is not a class
        else:
            if ormclasstype.lower() == 'group':
                ormclasstype = ormclasstype.lower()
                query_type_string = None
                ormclass = self.Group
            elif ormclasstype.lower() == 'computer':
                ormclasstype = ormclasstype.lower()
                query_type_string = None
                ormclass = self.Computer
            elif ormclasstype.lower() == 'user':
                ormclasstype = ormclasstype.lower()
                query_type_string = None
                ormclass = self.User
            else:
                # At this point, it has to be a node.
                # The only valid string at this point is a string
                # that matches exactly the _plugin_type_string
                # of a node class
                from aiida.common.pluginloader import (
                        from_type_to_pluginclassname,
                        load_plugin
                    )
                ormclass = self.Node
                try:
                    pluginclassname = from_type_to_pluginclassname(ormclasstype)

                    # I want to check at this point if that is a valid class,
                    # so I use the load_plugin to load the plugin class
                    # and use the classes _plugin_type_string attribute
                    # In the future, assuming the user knows what he or she is doing
                    # we could remove that check
                    # The query_type_string we can get from
                    # the aiida.common.pluginloader function get_query_type_string
                    PluginClass = load_plugin(self.AiidaNode, 'aiida.orm', pluginclassname)
                except (DbContentError, MissingPluginError) as e:
                    raise InputValidationError(
                        "\nYou provide a vertice of the path with\n"
                        "type={}\n"
                        "But that string is not a valid type string\n"
                        "Exception raise during check\n"
                        "{}".format(ormclasstype, e)
                    )


                ormclasstype = PluginClass._plugin_type_string
                query_type_string = PluginClass._query_type_string

        return ormclass, ormclasstype, query_type_string

    def _get_autotag(self, ormclasstype):
        basetag = self._get_tag_from_type(ormclasstype)
        tags_used = self._tag_to_alias_map.keys()
        for i in range(1, 100):
            tag = '{}_{}'.format(basetag, i)
            if tag not in tags_used:
                return tag


    def _get_tag_from_type(self, ormclasstype):
        """
        Assign a tag to the given
        vertice of a path, based mainly on the type
        *   data.structure.StructureData -> StructureData
        *   data.structure.StructureData. -> StructureData
        *   calculation.job.quantumespresso.pw.PwCalculation. -. PwCalculation
        *   node.Node. -> Node
        *   Node -> Node
        *   computer -> computer
        """
        return ormclasstype.rstrip('.').split('.')[-1]


    def append(self, cls=None, type=None, tag=None,
                autotag=False, filters=None, project=None, subclassing=True,
                edge_tag=None, edge_filters=None, edge_project=None, **kwargs
        ):
        """
        Any iterative procedure to build the path for a graph query
        needs to invoke this method to append to the path.

        :param cls: The Aiida-class (or backend-class) defining the appended vertice
        :param type: The type of the class, if cls is not given
        :param tag:
            A unique tag. If none is given, will take the classname.
            See keyword autotag to achieve unique tag.
        :param filters:
            Filters to apply for this vertice.
            See usage examples for details.
        :param autotag:
            Whether to search for a unique tag,
            (default **False**). If **True**, will find a unique tag.
            Cannot be set to **True** if tag is specified.
        :param subclassing:
            Whether to include subclasses of the given class
            (default **True**).
            E.g. Specifying JobCalculation will include PwCalculation

        A small usage example how this can be invoked::

            qb = QueryBuilder()             # Instantiating empty querybuilder instance
            qb.append(cls=StructureData)    # First item is StructureData node
            # The
            # next node in the path is a PwCalculation, with
            # the structure joined as an input
            qb.append(
                cls=PwCalculation,
                output_of=StructureData
            )

        :returns: self
        """
        ######################## INPUT CHECKS ##########################
        # This function can be called by users, so I am checking the
        # input now.
        # First of all, let's make sure the specified
        # the class or the type (not both)
        if cls and type:
            raise InputValidationError(
                    "\n\n\n"
                    "You cannot specify both a \n"
                    "class ({})\n"
                    "and a type ({})\n\n"
                    "".format(cls, type)
                )

        if not (cls or type):
            raise InputValidationError(
                    "\n\n"
                    "You need to specify either a class or a type"
                    "\n\n"
                )

        # Let's check if it is a valid class or type
        if cls:
            if not inspect_isclass(cls):
                raise InputValidationError(
                    "\n\n"
                    "{} was passed with kw 'cls', but is not a class"
                    "\n\n".format(cls)
                )
        elif type:
            if not isinstance(type, basestring):
                raise InputValidationError(
                    "\n\n\n"
                    "{} was passed as type, but is not a string"
                    "\n\n\n".format(type)
                )

        if 'link_tag' in kwargs:
            raise DeprecationWarning("link_tag is deprecated, use edge_tag instead")
        ormclass, ormclasstype, query_type_string = self._get_ormclass(cls, type)
        ############################### TAG #################################
        # Let's get a tag
        user_defined_tag = False
        label = kwargs.pop('label', None)
        if label is not None:
            if tag is None:
                warnings.warn(
                    "\nUse of the keyword 'label' will be deprecated soon\n"
                    "Please use 'tag' instead\n",
                    DeprecationWarning,
                )
                tag = label
            else:
                raise InputValidationError("Both label and tag specified")

        if tag:
            if self._EDGE_TAG_DELIM in tag:
                raise InputValidationError(
                    "tag cannot contain {}\n"
                    "since this is used as a delimiter for links"
                    "".format(self._EDGE_TAG_DELIM)
                )
            tag = tag
            user_defined_tag = True
        elif autotag:
            tag = self._get_autotag(ormclasstype)
        else:
            tag = self._get_tag_from_type(ormclasstype)
        # Check if the tag is not yet used:
        if tag in self._tag_to_alias_map.keys():
            if user_defined_tag:
                raise InputValidationError(
                    "\n"
                    "This tag ({}) is already in use\n"
                    "\n".format(tag)
                )
            else:
                raise InputValidationError(
                    "\n"
                    "You did not specify a tag, so I am making one myself\n"
                    "based on the class/type you gave me\n"
                    "The tag that I made ({}) is already in use\n"
                    "please specify a tag or set autotag to true"
                    "".format(tag)
                )

        ################ TAG MAPPING #################################
        # Let's fill the cls_to_tag_map so that one can specify
        # this vertice in a joining specification later
        # First this only makes sense if a class was specified:
        if cls:
            if cls in self._cls_to_tag_map.keys():
                # In this case, this class already stands for another
                # tag that was used before.
                # This means that the first tag will be the correct
                # one. This is dangerous and maybe should be avoided in
                # the future
                pass

            else:
                self._cls_to_tag_map[cls] = tag
            # TODO check with duplicate classes


        ######################## ALIASING ##############################
        alias = aliased(ormclass)
        self._aliased_path.append(alias)
        self._tag_to_alias_map[tag] = alias



        ################# FILTERS ######################################

        self._filters[tag] = {}

        # I have to add a filter on column type.
        # This so far only is necessary for AiidaNodes
        # GROUPS?
        if query_type_string is not None:
            self._add_type_filter(tag, query_type_string, subclassing)
        # The order has to be first _add_type_filter and then add_filter.
        # If the user adds a query on the type column, it overwrites what I did

        #if the user specified a filter, add it:
        if filters is not None:
            self.add_filter(tag, filters)


        ##################### PROJECTIONS ##############################
        self._projections[tag] = []

        if project is not None:
            self.add_projection(tag, project)




        ################## JOINING #####################################

        # Get the functions that are implemented:
        spec_to_function_map = self._get_function_map().keys()


        joining_keyword = kwargs.pop('joining_keyword', None)
        joining_value = kwargs.pop('joining_value', None)
        #~ reverse_linklabel = kwargs.pop('reverse_linklabel', None)


        for key, val in kwargs.items():
            if key not in spec_to_function_map:
                raise InputValidationError(
                        "\n\n\n"
                        "{} is not a valid keyword "
                        "for joining specification\n"
                        "Valid keywords are:\n"
                        "{}\n\n\n".format(
                                key,
                                spec_to_function_map+[
                                    'cls', 'type', 'tag',
                                    'autotag', 'filters', 'project'
                                ]
                            )
                    )
            elif joining_keyword:
                raise InputValidationError(
                        "\n\n\n"
                        "You already specified joining specification {}\n"
                        "But you now also want to specify {}"
                        "\n\n\n".format(joining_keyword, key)
                    )
            else:
                joining_keyword = key
                joining_value = self._get_tag_from_specification(val)
        # the default is that this vertice is 'output_of' the previous one
        if joining_keyword is None and len(self._path)>0:
            joining_keyword = 'output_of'
            joining_value = self._path[-1]['tag']

        if joining_keyword == 'direction':
            if not isinstance(joining_value, int):
                raise InputValidationError("direction=n expects n to be an integer")
            try:
                if joining_value < 0:
                    joining_keyword = 'input_of'
                elif joining_value > 0:
                    joining_keyword = 'output_of'
                else:
                    raise InputValidationError("direction=0 is not valid")
                joining_value = self._path[-abs(joining_value)]['tag']
            except IndexError as e:
                raise InputValidationError(
                    "You have specified a non-existent entity with\n"
                    "direction={}\n"
                    "{}\n".format(joining_value, e.message)
                )

        ############################# EDGES #################################
        # See if this requires a link:
        aliased_edge = None
        if len(self._path) > 0:
            if joining_keyword in ('input_of', 'output_of'):
                aliased_edge = aliased(self.Link)
            elif joining_keyword in ('ancestor_of', 'descendant_of'):
                aliased_edge = aliased(self.Path)

            if aliased_edge is not None:

                # Ok, so here we are joining through a m2m relationship,
                # e.g. input or output.
                # This means that the user might want to query by that edge!

                if edge_tag is None:
                    edge_destination_tag = self._get_tag_from_specification(joining_value)
                    edge_tag = edge_destination_tag + self._EDGE_TAG_DELIM + tag
                else:
                    if edge_tag in self._tag_to_alias_map.keys():
                        raise InputValidationError(
                            "The tag {} is already in use".format(edge_tag)
                        )

                self._tag_to_alias_map[edge_tag] = aliased_edge

                # Filters on links:
                self._filters[edge_tag] = {}
                if edge_filters is not None:
                    self.add_filter(edge_tag, edge_filters)

                # Projections on links
                self._projections[edge_tag] = {}
                if edge_project is not None:
                    self.add_projection(edge_tag, edge_project)


        ################### EXTENDING THE PATH #################################


        path_extension = dict(
                type=ormclasstype, tag=tag, joining_keyword=joining_keyword,
                joining_value=joining_value
            )
        if aliased_edge is not None:
            path_extension.update(dict(edge_tag=edge_tag))
            #~ if reverse_linktag is not None:
                #~ path_extension.update(dict(reverse_linktag=reverse_linktag))

        self._path.append(path_extension)
        return self

    def order_by(self, order_by):
        """
        Set the entity to order by

        :param order_by:
            This is a list of items, where each item is a dictionary specifies
            what to sort for an entity

        In each dictionary in that list, keys represent valid tags of
        entities (tables), and values are list of columns.

        Usage::

            #Sorting by id (ascending):
            qb = QueryBuilder()
            qb.append(Node, tag='node')
            qb.order_by({'node':['id']})

            # or
            #Sorting by id (ascending):
            qb = QueryBuilder()
            qb.append(Node, tag='node')
            qb.order_by({'node':[{'id':{'order':'asc'}}]})

            # for descending order:
            qb = QueryBuilder()
            qb.append(Node, tag='node')
            qb.order_by({'node':[{'id':{'order':'desc'}}]})

            # or (shorter)
            qb = QueryBuilder()
            qb.append(Node, tag='node')
            qb.order_by({'node':[{'id':'desc'}]})


        """

        self._order_by = []
        allowed_keys = ('cast', 'order')
        possible_orders = ('asc', 'desc')

        if not isinstance(order_by, (list, tuple)):
            order_by = [order_by]

        for order_spec in order_by:
            if not isinstance(order_spec, dict):
                    raise InputValidationError(
                        "Invalid input for order_by statement: {}\n"
                        "I am expecting a dictionary ORMClass,"
                        "[columns to sort]"
                        "".format(order_spec)
                    )
            _order_spec = {}
            for tagspec,items_to_order_by in order_spec.items():
                if not isinstance(items_to_order_by, (tuple, list)):
                    items_to_order_by = [items_to_order_by]
                tag = self._get_tag_from_specification(tagspec)
                _order_spec[tag] = []
                for item_to_order_by in items_to_order_by:
                    if isinstance(item_to_order_by, basestring):
                        item_to_order_by = {item_to_order_by:{}}
                    elif isinstance(item_to_order_by, dict):
                        pass
                    else:
                        raise InputValidationError(
                            "Cannot deal with input to order_by {}\n"
                            "of type{}"
                            "\n".format(item_to_order_by, type(item_to_order_by))
                        )
                    for entityname, orderspec in item_to_order_by.items():
                        # if somebody specifies eg {'node':{'id':'asc'}}
                        # tranform to {'node':{'id':{'order':'asc'}}}

                        if isinstance(orderspec, basestring):
                            this_order_spec = {'order':orderspec}
                        elif isinstance(orderspec, dict):
                            this_order_spec = orderspec
                        else:
                            raise InputValidationError(
                                "I was expecting a string or a dictionary\n"
                                "You provided {} {}\n"
                                "".format(type(orderspec), orderspec)
                            )
                        for key in this_order_spec.keys():
                            if key not in allowed_keys:
                                raise InputValidationError(
                                    "The allowed key for an order specification\n"
                                    "are {}\n"
                                    "{} is not valid\n"
                                    "".format(', '.join(allowed_keys), k)
                                )
                        this_order_spec['order'] = this_order_spec.get('order', 'asc')
                        if this_order_spec['order'] not in possible_orders:
                            raise InputValidationError(
                                "You gave {} as an order parameters,\n"
                                "but it is not a valid order parameter\n"
                                "Valid orders are: {}\n"
                                "".format(this_order_spec['order'], possible_orders)
                            )
                        item_to_order_by[entityname] = this_order_spec

                    _order_spec[tag].append(item_to_order_by)

            self._order_by.append(_order_spec)
        return self

    def add_filter(self, tagspec, filter_spec):
        """
        Adding a filter to my filters.

        :param tagspec:
            The tag, which has to exist already as a key
            in self._filters
        :param filter_spec:
            The specifications for the filter, has to be a dictionary
        """


        if not isinstance(filter_spec, dict):
            raise InputValidationError(
                    "Filters have to be passed as dictionaries"
                )

        tag = self._get_tag_from_specification(tagspec)
        self._filters[tag].update(filter_spec)

    def _add_type_filter(
        self, tagspec, query_type_string,
        ormclasstype, subclassing=True):
        """
        Add a filter on the type based on the query_type_string
        """
        tag = self._get_tag_from_specification(tagspec)

        if subclassing:
            node_type_flt = {'like':'{}%'.format(query_type_string)}
        else:
            node_type_flt = {'==':ormclasstype}

        self.add_filter(tagspec, {'type':node_type_flt})

    def add_projection(self, tag_spec, projection_spec):
        """
        Adds a projection

        :param tag_spec: A valid specification for a tag
        :param projection_spec:
            The specification for the projection.
            A projection is a list of dictionaries, with each dictionary
            containing key-value pairs where the key is database entity
            (e.g. a column / an attribute) and the value is (optional)
            additional information on how to process this database entity.

        If the given *projection_spec* is not a list, it will be expanded to
        a list.
        If the listitems are not dictionaries, but strings (No additional
        processing of the projected results desired), they will be expanded to
        dictionaries.

        Usage::

            qb = QueryBuilder()
            qb.append(StructureData, tag='struc')

            # Will project the uuid and the kinds
            qb.add_projection('struc', ['uuid', 'attributes.kinds'])
        """
        tag = self._get_tag_from_specification(tag_spec)
        _projections = []
        if not isinstance(projection_spec, (list, tuple)):
            projection_spec = [projection_spec]
        for projection in projection_spec:
            if isinstance(projection, dict):
                _thisprojection = projection
            elif isinstance(projection, basestring):
                _thisprojection = {projection:{}}
            else:
                raise InputValidationError(
                    "Cannot deal with projection specification {}\n"
                    "".format(projection)
                )
            for p,spec in _thisprojection.items():
                if not isinstance(spec, dict):
                    raise InputValidationError(
                        "\nThe value of a key-value pair in a projection\n"
                        "has to be a dictionary\n"
                        "You gave: {}\n"
                        "".format(spec)
                    )

                for key, val in spec.items():
                    if key not in self._VALID_PROJECTION_KEYS:
                        raise InputValidationError(
                            "{} is not a valid key {}".format(
                                key, self._VALID_PROJECTION_KEYS)
                        )
                        if not isinstance(val, basestring):
                            raise InputValidationError(
                                "{} has to be a string".format(val)
                            )
            _projections.append(_thisprojection)
        self._projections[tag] = _projections


    def _get_projectable_entity(self, alias, column_name, attrpath, **entityspec):

        if len(attrpath) or column_name in ('attributes', 'extras'):

            entity = self._get_projectable_attribute(
                        alias, column_name, attrpath, **entityspec
                    )
        else:
            entity = self._get_column(column_name, alias)
        return entity

    def _add_to_projections(self, alias, projectable_entity_name, cast=None, func=None):
        """
        :param alias:
            A instance of *sqlalchemy.orm.util.AliasedClass*, alias for an ormclass
        :param projectable_entity_name:
            User specification of what to project.
            Appends to query's entities what the user wants to project
            (have returned by the query)

        """
        column_name = projectable_entity_name.split('.')[0]
        attr_key = projectable_entity_name.split('.')[1:]


        if column_name == '*':
            if func is not None:
                raise InputValidationError(
                        "Very sorry, but functions on the aliased class\n"
                        "(You specified '*')\n"
                        "will not work!\n"
                        "I suggest you apply functions on a column, e.g. ('id')\n"
                    )
            self._query = self._query.add_entity(alias)
        else:
            entity_to_project = self._get_projectable_entity(
                    alias, column_name, attr_key,
                    cast=cast
                )
            if func is None:
                pass
            elif func == 'max':
                entity_to_project = sa_func.max(entity_to_project)
            elif func == 'min':
                entity_to_project = sa_func.max(entity_to_project)
            elif func == 'count':
                entity_to_project = sa_func.count(entity_to_project)
            else:
                raise InputValidationError(
                        "\nInvalid function specification {}".format(func)
                    )
            self._query =  self._query.add_columns(entity_to_project)




    def _build_projections(self, tag, items_to_project=None):

        if items_to_project is None:
            items_to_project = self._projections.get(tag, [])

        # Return here if there is nothing to project,
        # reduces number of key in return dictionary
        if not items_to_project:
            return

        alias = self._tag_to_alias_map[tag]

        self.tag_to_projected_entity_dict[tag] = {}

        for projectable_spec in items_to_project:
            for projectable_entity_name, extraspec in projectable_spec.items():
                if projectable_entity_name == '**':
                    # Need to expand
                    entity_names = [str(c).replace(alias.__table__.name+'.','') for c in alias.__table__.columns]
                    #~ for s in ('attributes', 'extras'):
                        #~ try:
                            #~ entity_names.remove(s)
                        #~ except ValueError:
                            #~ pass
                else:
                    entity_names = [projectable_entity_name]
                for entity_name in entity_names:
                    self._add_to_projections(
                            alias, entity_name, **extraspec
                        )

                    self.tag_to_projected_entity_dict[tag][
                            entity_name
                        ] = self.nr_of_projections
                    self.nr_of_projections += 1


    def _get_tag_from_specification(self, specification):
        if isinstance(specification, basestring):
            if specification in self._tag_to_alias_map.keys():
                tag = specification
            else:
                raise InputValidationError(
                        "tag {} is not among my known tags\n"
                        "   My tags are: {}"
                        "\n\n".format(
                                specification, self._tag_to_alias_map.keys()
                            )
                    )
        else:
            if specification in self._cls_to_tag_map.keys():
                tag = self._cls_to_tag_map[specification]
            else:
                raise InputValidationError(
                    "\nYou specified as a class for which I have to find a tag\n"
                    "The classes that I can do this for are:{}\n"
                    "The tags I have are: {}\n"
                    "\n".format(
                        specification, self._cls_to_tag_map.keys(),
                        self._tag_to_alias_map.keys()
                    )
                )
        return tag

    def limit(self, limit):
        """
        Set the limit (nr of rows to return)

        :param int limit: integers of nr of rows to return
        """

        if limit is not None:
            if not isinstance(limit, int):
                raise InputValidationError("limit has to be an integer")
        self._limit = limit
        return self

    def offset(self, offset):
        """
        Set the offset. If offset is set, that many rows are skipped before returning.
        *offset* = 0 is the same as omitting setting the offset.
        If both offset and limit appear,
        then *offset* rows are skipped before starting to count the *limit* rows
        that are returned.

        :param int offset: integers of nr of rows to skip
        """
        if offset is not None:
            if not isinstance(offset, int):
                raise InputValidationError(
                    "offset has to be an integer"
                )
        self._offset = offset
        return self


    @staticmethod
    @abstractmethod
    def _get_session():
        pass

    @staticmethod
    def _get_filter_expr_from_column(operator, value, column):

        if not isinstance(column, (Cast, InstrumentedAttribute)):
            raise TypeError(
                'column ({}) {} is not a valid column'.format(
                    type(column), column
                )
            )
        database_entity = column
        if operator == '==':
            expr = database_entity == value
        elif operator == '>':
            expr = database_entity > value
        elif operator == '<':
            expr = database_entity < value
        elif operator == '>=':
            expr = database_entity >= value
        elif operator == '<=':
            expr = database_entity <= value
        elif operator == 'like':
            expr = database_entity.like(value)
        elif operator == 'ilike':
            expr = database_entity.ilike(value)
        elif operator == 'in':
            expr = database_entity.in_(value)
        else:
            raise InputValidationError(
                'Unknown operator {} for filters on columns'.format(operator)
            )
        return expr


    @classmethod
    def _get_filter_expr(
            cls, operator, value, attr_key, is_attribute,
            alias=None, column=None, column_name=None
        ):
        """
        Applies a filter on the alias given.
        Expects the alias of the ORM-class on which to filter, and filter_spec.
        Filter_spec contains the specification on the filter.
        Expects:

        :param operator: The operator to apply, see below for further details
        :param value:
            The value for the right side of the expression,
            the value you want to compare with.

        :param path: The path leading to the value

        :param attr_key: Boolean, whether the value is in a json-column,
            or in an attribute like table.


        Implemented and valid operators:

        *   for any type:
            *   ==  (compare single value, eg: '==':5.0)
            *   in    (compare whether in list, eg: 'in':[5, 6, 34]
        *  for floats and integers:
            *   >
            *   <
            *   <=
            *   >=
        *  for strings:
            *   like  (case - sensitive), for example
                'like':'node.calc.%'  will match node.calc.relax and
                node.calc.RELAX and node.calc. but
                not node.CALC.relax
            *   ilike (case - unsensitive)
                will also match node.CaLc.relax in the above example

            .. note::
                The character % is a reserved special character in SQL,
                and acts as a wildcard. If you specifically
                want to capture a ``%`` in the string, use: ``_%``

        *   for arrays and dictionaries (only for the
            SQLAlchemy implementation):

            *   contains: pass a list with all the items that
                the array should contain, or that should be among
                the keys, eg: 'contains': ['N', 'H'])
            *   has_key: pass an element that the list has to contain
                or that has to be a key, eg: 'has_key':'N')

        *  for arrays only (SQLAlchemy version):
            *   of_length
            *   longer
            *   shorter

        All the above filters invoke a negation of the
        expression if preceded by **~**::

            # first example:
            filter_spec = {
                'name' : {
                    '~in':[
                        'halle',
                        'lujah'
                    ]
                } # Name not 'halle' or 'lujah'
            }

            # second example:
            filter_spec =  {
                'id' : {
                    '~==': 2
                }
            } # id is not 2
        """


        expr = None
        if operator.startswith('~'):
            negation = True
            operator = operator.lstrip('~')
        elif operator.startswith('!'):
            negation = True
            operator = operator.lstrip('!')
        else:
            negation = False
        if operator in ('longer', 'shorter', 'of_length'):
            if not isinstance(value, int):
                raise InputValidationError(
                    "You have to give an integer when comparing to a length"
                )
        elif operator in ('like', 'ilike'):
            if not isinstance(value, basestring):
                raise InputValidationError(
                    "Value for operator {} has to be a string (you gave {})"
                    "".format(operator, value)
                )

        elif operator == 'in':
            value_type_set = set([type(i) for i in value])
            if len(value_type_set) > 1:
                raise InputValidationError(
                        '{}  contains more than one type'.format(value)
                    )
            elif len(value_type_set) == 0:
                raise InputValidationError(
                        '{}  contains is an empty list'.format(value)
                    )
        elif operator in ('and', 'or'):
            expressions_for_this_path = []
            for filter_operation_dict in value:
                for newoperator, newvalue in filter_operation_dict.items():
                    expressions_for_this_path.append(
                            cls._get_filter_expr(
                                    newoperator, newvalue,
                                    attr_key=attr_key, is_attribute=is_attribute,
                                    alias=alias, column=column,
                                    column_name=column_name
                                )
                        )
            if operator == 'and':
                expr = and_(*expressions_for_this_path)
            elif operator == 'or':
                 expr = or_(*expressions_for_this_path)

        if expr is None:
            if is_attribute:
                expr = cls._get_filter_expr_from_attributes(
                        operator, value, attr_key,
                        column=column, column_name=column_name, alias=alias
                    )
            else:
                if column is None:
                    if (alias is None) and (column_name is None):
                        raise Exception(
                            "I need to get the column but do not know \n"
                            "the alias and the column name"
                        )
                    column = cls._get_column(column_name, alias)
                expr = cls._get_filter_expr_from_column(operator, value, column)
        if negation:
            return not_(expr)
        return expr



    def _build_filters(self, alias, filter_spec):
        """
        Recurse through the filter specification and apply filter operations.

        :param alias: The alias of the ORM class the filter will be applied on
        :param filter_spec: the specification as given by the queryhelp

        :returns: an instance of *sqlalchemy.sql.elements.BinaryExpression*.
        """
        expressions = []
        for path_spec, filter_operation_dict in filter_spec.items():
            if path_spec in  ('and', 'or', '~or', '~and', '!and', '!or'):
                subexpressions = [
                    self._build_filters(alias, sub_filter_spec)
                    for sub_filter_spec in filter_operation_dict
                ]
                if path_spec == 'and':
                    expressions.append(and_(*subexpressions))
                elif path_spec == 'or':
                    expressions.append(or_(*subexpressions))
                elif path_spec in ('~and', '!and'):
                    expressions.append(not_(and_(*subexpressions)))
                elif path_spec in ('~or', '!or'):
                    expressions.append(not_(or_(*subexpressions)))
            else:
                column_name = path_spec.split('.')[0]

                attr_key = path_spec.split('.')[1:]
                is_attribute = (
                    attr_key or
                    column_name in ('attributes', 'extras')
                )
                try:
                    column = self._get_column(column_name, alias)
                except InputValidationError as e:
                    if is_attribute:
                        column = None
                    else:
                        raise e
                #~ is_attribute = bool(attr_key)
                if not isinstance(filter_operation_dict, dict):
                    filter_operation_dict = {'==':filter_operation_dict}
                [
                    expressions.append(
                        self._get_filter_expr(
                            operator, value, attr_key,
                            is_attribute=is_attribute,
                            column=column, column_name=column_name,
                            alias=alias
                        )
                    )
                    for operator, value
                    in filter_operation_dict.items()
                ]
        return and_(*expressions)

    #~ @abstractmethod
    #~ def _build_filters(self, alias, filter_spec):
        #~ pass
    @staticmethod
    def _check_dbentities(entities_cls_joined, entities_cls_to_join, relationship):
        """
        :param list entities_cls_joined:
            A list (tuple) of the aliased class passed as joined_entity and
            the ormclass that was expected
        :param list entities_cls_joined:
            A list (tuple) of the aliased class passed as entity_to_join and
            the ormclass that was expected
        :param str relationship:
            The relationship between the two entities to make the Exception
            comprehensible
        """
        for entity, cls in (entities_cls_joined, entities_cls_to_join):

            if not issubclass(entity._sa_class_manager.class_, cls):
                raise InputValidationError(
                    "\nYou are attempting to join {} as '{}' of {}\n"
                    "This failed because you passed:\n"
                    " - {} as entity joined (expected {})\n"
                    " - {} as entity to join (expected {})\n"
                    "\n".format(
                        entities_cls_joined[0],
                        relationship,
                        entities_cls_to_join[0],
                        entities_cls_joined[0]._sa_class_manager.class_,
                        entities_cls_joined[1],
                        entities_cls_to_join[0]._sa_class_manager.class_,
                        entities_cls_to_join[1],
                    )
                )


    def _join_slaves(self, joined_entity, entity_to_join):
        raise NotImplementedError(
                "Master - slave relationships are not implemented"
            )
        #~ call = aliased(Call)
        #~ self._query = self._query.join(call,  call.caller_id == joined_entity.id)
        #~ self._query = self._query.join(
                #~ entity_to_join,
                #~ call.called_id == entity_to_join.id
            #~ )

    def _join_masters(self, joined_entity, entity_to_join):
        raise NotImplementedError(
                "Master - slave relationships are not implemented"
            )
        #~ call = aliased(Call)
        #~ self._query = self._query.join(call,  call.called_id == joined_entity.id)
        #~ self._query = self._query.join(
                #~ entity_to_join,
                #~ call.caller_id == entity_to_join.id
            #~ )

    def _join_outputs(self, joined_entity, entity_to_join, aliased_edge):
        """
        :param joined_entity: The (aliased) ORMclass that is an input
        :param entity_to_join: The (aliased) ORMClass that is an output.

        **joined_entity** and **entity_to_join** are joined with a link
        from **joined_entity** as input to **enitity_to_join** as output
        (**enitity_to_join** is an *output_of* **joined_entity**)
        """
        self._check_dbentities(
                (joined_entity, self.Node),
                (entity_to_join, self.Node),
                'output_of'
            )
        self._query = self._query.join(
                aliased_edge,
                aliased_edge.input_id == joined_entity.id
        ).join(
                entity_to_join,
                aliased_edge.output_id == entity_to_join.id
        )

    def _join_inputs(self, joined_entity, entity_to_join, aliased_edge):
        """
        :param joined_entity: The (aliased) ORMclass that is an output
        :param entity_to_join: The (aliased) ORMClass that is an input.

        **joined_entity** and **entity_to_join** are joined with a link
        from **joined_entity** as output to **enitity_to_join** as input
        (**enitity_to_join** is an *input_of* **joined_entity**)
        """
        self._check_dbentities(
                (joined_entity, self.Node),
                (entity_to_join, self.Node),
                'input_of'
            )
        self._query = self._query.join(
                aliased_edge,
                aliased_edge.output_id == joined_entity.id
        ).join(
                entity_to_join,
                aliased_edge.input_id == entity_to_join.id
        )

    def _join_descendants(self, joined_entity, entity_to_join, aliased_path):
        """
        :param joined_entity: The (aliased) ORMclass that is an ancestor
        :param entity_to_join: The (aliased) ORMClass that is a descendant.
        :param aliased_path: An aliased instance of DbPath

        **joined_entity** and **entity_to_join** are
        joined via the DbPath table.
        from **joined_entity** as parent to **enitity_to_join** as child
        (**enitity_to_join** is a *descendant_of* **joined_entity**)
        """
        self._check_dbentities(
                (joined_entity, self.Node),
                (entity_to_join, self.Node),
                'descendant_of'
            )
        
        self._query = self._query.join(
                aliased_path,
                aliased_path.parent_id == joined_entity.id
        ).join(
                entity_to_join,
                aliased_path.child_id == entity_to_join.id
        )

    def _join_ancestors(self, joined_entity, entity_to_join, aliased_path):
        """
        :param joined_entity: The (aliased) ORMclass that is a descendant
        :param entity_to_join: The (aliased) ORMClass that is an ancestor.
        :param aliased_path: An aliased instance of DbPath

        **joined_entity** and **entity_to_join**
        are joined via the DbPath table.
        from **joined_entity** as child to **enitity_to_join** as parent
        (**enitity_to_join** is an *ancestor_of* **joined_entity**)
        """
        self._check_dbentities(
                (joined_entity, self.Node),
                (entity_to_join, self.Node),
                'ancestor_of'
            )
        #~ aliased_path = aliased(self.Path)
        self._query = self._query.join(
                aliased_path,
                aliased_path.child_id == joined_entity.id
        ).join(
                entity_to_join,
                aliased_path.parent_id == entity_to_join.id
        )
    def _join_group_members(self, joined_entity, entity_to_join):
        """
        :param joined_entity:
            The (aliased) ORMclass that is
            a group in the database
        :param entity_to_join:
            The (aliased) ORMClass that is a node and member of the group

        **joined_entity** and **entity_to_join**
        are joined via the table_groups_nodes table.
        from **joined_entity** as group to **enitity_to_join** as node.
        (**enitity_to_join** is an *member_of* **joined_entity**)
        """
        self._check_dbentities(
                (joined_entity, self.Group),
                (entity_to_join, self.Node),
                'member_of'
            )
        aliased_group_nodes = aliased(self.table_groups_nodes)
        self._query = self._query.join(
                aliased_group_nodes,
                aliased_group_nodes.c.dbgroup_id == joined_entity.id
        ).join(
                entity_to_join,
                entity_to_join.id == aliased_group_nodes.c.dbnode_id
        )
    def _join_groups(self, joined_entity, entity_to_join):
        """
        :param joined_entity: The (aliased) node in the database
        :param entity_to_join: The (aliased) Group

        **joined_entity** and **entity_to_join** are
        joined via the table_groups_nodes table.
        from **joined_entity** as node to **enitity_to_join** as group.
        (**enitity_to_join** is an *group_of* **joined_entity**)
        """
        self._check_dbentities(
                (joined_entity, self.Node),
                (entity_to_join, self.Group),
                'group_of'
            )
        aliased_group_nodes = aliased(self.table_groups_nodes)
        self._query = self._query.join(
                aliased_group_nodes,
                aliased_group_nodes.c.dbnode_id == joined_entity.id
        ).join(
                entity_to_join,
                entity_to_join.id == aliased_group_nodes.c.dbgroup_id
        )
    def _join_creator_of(self, joined_entity, entity_to_join):
        """
        :param joined_entity: the aliased node
        :param entity_to_join: the aliased user to join to that node
        """
        self._check_dbentities(
                (joined_entity, self.Node),
                (entity_to_join, self.User),
                'creator_of'
            )
        self._query = self._query.join(
                entity_to_join,
                entity_to_join.id == joined_entity.user_id
            )
    def _join_created_by(self, joined_entity, entity_to_join):
        """
        :param joined_entity: the aliased user you want to join to
        :param entity_to_join: the (aliased) node or group in the DB to join with
        """
        self._check_dbentities(
                (joined_entity, self.User),
                (entity_to_join, self.Node),
                'created_by'
            )
        self._query = self._query.join(
                entity_to_join,
                entity_to_join.user_id == joined_entity.id
            )

    def _join_to_computer_used(self, joined_entity, entity_to_join):
        """
        :param joined_entity: the (aliased) computer entity
        :param entity_to_join: the (aliased) node entity

        """
        self._check_dbentities(
                (joined_entity, self.Computer),
                (entity_to_join, self.Node),
                'has_computer'
            )
        self._query = self._query.join(
                entity_to_join,
                entity_to_join.dbcomputer_id == joined_entity.id
        )

    def _join_computer(self, joined_entity, entity_to_join):
        """
        :param joined_entity: An entity that can use a computer (eg a node)
        :param entity_to_join: aliased dbcomputer entity


        """
        self._check_dbentities(
                (joined_entity, self.Node),
                (entity_to_join, self.Computer),
                'computer_of'
            )
        self._query = self._query.join(
                entity_to_join,
                joined_entity.dbcomputer_id == entity_to_join.id
        )

    def _get_function_map(self):
        d = {
                'input_of'  : self._join_inputs,
                'output_of' : self._join_outputs,
                'slave_of'  : self._join_slaves, # not implemented
                'master_of' : self._join_masters,# not implemented
                'ancestor_of': self._join_ancestors,
                'descendant_of': self._join_descendants,
                'direction' : None,
                'group_of'  : self._join_groups,
                'member_of' : self._join_group_members,
                'has_computer':self._join_to_computer_used,
                'computer_of':self._join_computer,
                'created_by' : self._join_created_by,
                'creator_of' : self._join_creator_of,
        }
        return d
    def _get_connecting_node(
            self, index,
            joining_keyword=None, joining_value=None, **kwargs
        ):
        """
        :param querydict:
            A dictionary specifying how the current node
            is linked to other nodes.
        :param index: Index of this node within the path specification

        Valid (currently implemented) keys are:

        *   *input_of*
        *   *output_of*
        *   *descendant_of*
        *   *ancestor_of*
        *   *direction*
        *   *group_of*
        *   *member_of*
        *   *has_computer*
        *   *computer_of*
        *   *created_by*
        *   *creator_of*

        Future:

        *   *master_of*
        *   *slave_of*
        """

        if joining_keyword == 'direction':
            if joining_value > 0:
                returnval = self._aliased_path[index-joining_value], self._join_outputs
            elif joining_value < 0:
                returnval = self._aliased_path[index+joining_value], self._join_inputs
            else:
                raise Exception("Direction 0 is not valid")
        else:
            func = self._get_function_map()[joining_keyword]

            if isinstance(joining_value, int):
                returnval = (self._aliased_path[joining_value], func)
            elif isinstance(joining_value, str):
                try:
                    returnval = self._tag_to_alias_map[
                            self._get_tag_from_specification(joining_value)
                        ], func
                except KeyError:
                    raise InputValidationError(
                        'Key {} is unknown to the types I know about:\n'
                        '{}'.format(val, self._tag_to_alias_map.keys())
                    )
        return returnval

    def _get_json_compatible(self, inp):
        """

        :param inp:
            The input value that will be converted.
            Recurses into each value if **inp** is an iterable.
        """
        print inp
        if isinstance(inp, dict):
            for key, val in inp.items():
                inp[
                        self._get_json_compatible(key)
                ] = self._get_json_compatible(inp.pop(key))
        elif isinstance(inp, (list, tuple)):
            inp = [self._get_json_compatible(val) for val in inp]
        elif inspect_isclass(inp):
            if issubclass(inp, self.AiidaNode):
                return '.'.join(
                        inp._plugin_type_string.strip('.').split('.')[:-1]
                    )
            elif issubclass(inp, self.AiidaGroup):
                return 'group'
            else:
                raise InputValidationError
        else:
            try:
                inp = replacement_dict.get(inp, inp)
            except Exception as e:
                raise Exception("""
                Exception thrown: {}\n
                while replacing {}""".format(e, inp))
        return inp


    def get_json_compatible_queryhelp(self):
        """
        Makes the queryhelp a json - compatible  dictionary.
        In this way,the queryhelp can be stored in a node
        in the database and retrieved or shared.

        :returns: the json-compatible queryhelp

        All classes defined in the input are
        converted to strings specifying the type,
        for example:
        """
        from copy import deepcopy

        return deepcopy({
            'path'      :   self._path,
            'filters'   :   self._filters,
            'project'   :   self._projections,
            'order_by'  :   self._order_by,
            'limit'     :   self._limit,
            'offset'    :   self._offset,
        })

        #~ self._get_json_compatible()

    @staticmethod
    def _get_column(colname, alias):
        """
        Return the column for the projection, if the column name is specified.
        """

        if colname not in alias._sa_class_manager.mapper.c.keys():
            raise InputValidationError(
                "\n{} is not a column of {}\n".format(colname, alias)
            )
        return getattr(alias, colname)

    def _build_order(self, alias, entitytag, entityspec):

        column_name = entitytag.split('.')[0]
        attrpath = entitytag.split('.')[1:]
        if attrpath and 'cast' not in entityspec.keys():
            raise InputValidationError(
                "\n\n"
                "In order to project ({}), I have to cast the the values,\n"
                "but you have not specified the datatype to cast to\n"
                "You can do this with keyword 'cast'\n"
                "".format(entitytag)
            )


        entity = self._get_projectable_entity(alias, column_name, attrpath, **entityspec)
        order = entityspec.get('order', 'asc')
        if order == 'desc':
            entity = entity.desc()
        self._query = self._query.order_by(entity)


    def _build(self):
        """
        build the query and return a sqlalchemy.Query instance
        """

        # self.tags_location_dict is a dictionary that
        # maps the tag to its index in the list
        # this is basically the mapping between the count
        # of nodes traversed
        # and the tag used for that node
        self.tags_location_dict = {
                path['tag']:index
                for index, path
                in enumerate(self._path)
            }

        #Starting the query by receiving a session
        # Every subclass needs to have _get_session and give me the
        # right session
        firstalias = self._tag_to_alias_map[self._path[0]['tag']]
        self._query = self._get_session().query(firstalias)

        ######################### JOINS ################################

        for index, verticespec in  enumerate(self._path[1:], start=1):
            alias = self._tag_to_alias_map[verticespec['tag']]
            #looping through the queryhelp
            #~ if index:
                #There is nothing to join if that is the first table
            toconnectwith, connection_func = self._get_connecting_node(
                    index, **verticespec
                )
            edge_tag = verticespec.get('edge_tag', None)
            if edge_tag is None:
                connection_func(toconnectwith, alias)
            else:
                aliased_edge = self._tag_to_alias_map[edge_tag]
                connection_func(toconnectwith, alias, aliased_edge)

        ######################### FILTERS ##############################

        for tag, filter_specs in self._filters.items():
            try:
                alias = self._tag_to_alias_map[tag]
            except KeyError:
                # TODO Check KeyError before?
                raise InputValidationError(
                    ' You looked for tag {} among the alias list\n'
                    'The tags I know are:\n{}'
                    ''.format(tag, self._tag_to_alias_map.keys())
                )
            self._query = self._query.filter(
                    self._build_filters(alias, filter_specs)
                )

        ######################### PROJECTIONS ##########################
        # first clear the entities in the case the first item in the
        # path was not meant to be projected
        # attribute of Query instance storing entities to project:

        # Will be later set to this list:
        entities = []
        # Mapping between enitites and the tag used/ given by user:
        self.tag_to_projected_entity_dict = {}


        self.nr_of_projections = 0

        if not any(self._projections.values()):
            # If user has not set projection,
            # I will simply project the last item specified!
            # Don't change, path traversal querying
            # relies on this behavior!
            self._build_projections(self._path[-1]['tag'], items_to_project=[{'*':{}}])
        else:
            for vertice in self._path:
                self._build_projections(vertice['tag'])


        ##################### LINK-PROJECTIONS #########################

        for vertice in self._path:
            edge_tag = vertice.get('edge_tag', None)
            if edge_tag is not None:
                self._build_projections(edge_tag)

            #~ linktag = vertice.get('reverse_linktag', None)
            #~ if linktag is not None:
                #~ self._build_projections(linktag)

        ######################### ORDER ################################
        for order_spec in self._order_by:
            for tag, entities in order_spec.items():
                alias = self._tag_to_alias_map[tag]
                for entitydict in entities:
                    for entitytag, entityspec in entitydict.items():
                        self._build_order(alias, entitytag, entityspec)

        ######################### LIMIT ################################
        if self._limit is not None:
            self._query = self._query.limit(self._limit)

        ######################## OFFSET ################################
        if self._offset is not None:
            self._query = self._query.offset(self._offset)

        ################ LAST BUT NOT LEAST ############################
        #pop the entity that I added to start the query
        self._query._entities.pop(0)

        # Make a list that helps the projection postprocessing
        self._attrkeys_as_in_sql_result = {
            index_in_sql_result:attrkey
            for tag, projected_entities_dict
            in self.tag_to_projected_entity_dict.items()
            for attrkey, index_in_sql_result
            in projected_entities_dict.items()
        }

        if self.nr_of_projections > len(self._attrkeys_as_in_sql_result):
            raise InputValidationError(
                    "\nYou are projecting the same key\n"
                    "multiple times within the same node"
                )
        ######################### DONE #################################

        return self._query

    def except_if_input_to(self, calc_class):
        """
        Makes counterquery based on the own path, only selecting
        entries that have been input to *calc_class*

        :param calc_class: The calculation class to check against

        :returns: self
        """
        def build_counterquery(calc_class):
            if issubclass(calc_class, self.Node):
                orm_calc_class = calc_class
                type_spec = None
            elif issubclass(calc_class, self.AiidaNode):
                orm_calc_class = self.Node
                type_spec = calc_class._plugin_type_string
            else:
                raise Exception(
                    'You have given me {}\n'
                    'of type {}\n'
                    "and I don't know what to do with that"
                    ''.format(calc_class, type(calc_class))
                )

            input_alias_list = []
            for node in self._path:
                tag = node['tag']
                requested_cols = [
                        key

                        for item in self._projections[tag]
                        for key in item.keys()
                    ]
                if '*' in requested_cols:
                    input_alias_list.append(aliased(self._tag_to_alias_map[tag]))


            counterquery = self._get_session().query(orm_calc_class)
            if type_spec:
                counterquery = counterquery.filter(orm_calc_class.type == type_spec)
            for alias in input_alias_list:

                link = aliased(self.Link)
                counterquery = counterquery.join(
                    link,
                    orm_calc_class.id == link.output_id
                ).join(
                    alias,
                    alias.id == link.input_id)
                counterquery = counterquery.add_entity(alias)
            counterquery._entities.pop(0)
            return counterquery
        self._query = self.get_query()
        self._query = self._query.except_(build_counterquery(calc_class))
        return self


    def get_aliases(self):
        """
        :returns: the list of aliases
        """
        return self._aliased_path

    def get_alias(self, tag):
        """
        In order to continue a query by the user, this utility function
        returns the aliased ormclasses.

        :param tag: The tag for a vertice in the path
        :returns: the alias given for that vertice
        """
        tag = self._get_tag_from_specification(tag)
        return self._tag_to_alias_map[tag]



    def get_query(self):
        """

        Checks if the query instance is still valid by hashing the queryhelp.
        If not invokes :func:`QueryBuilderBase._build`.

        :returns: an instance of sqlalchemy.orm.Query

        """
        # Need_to_build is True by default.
        # It describes whether the current query
        # which is an attribute _query of this instance is still valid
        # The queryhelp_hash is used to determine
        # whether the query is still valid

        queryhelp_hash = make_hash(self.get_json_compatible_queryhelp())
        # if self._hash (which is None if this function has not been invoked
        # and is a string (hash) if it has) is the same as the queryhelp
        # I can use the query again:
        # If the query was injected I never build:
        if self._injected:
            need_to_build = False
        elif self._hash == queryhelp_hash:
            need_to_build = False
        else:
            need_to_build = True

        if need_to_build:
            query = self._build()
            self._hash = queryhelp_hash
        else:
            try:
                query = self._query
            except AttributeError:
                warnings.warn(
                    "AttributeError thrown even though I should\n"
                    "have _query as an attribute"
                )
                query = self._build()
                self._hash = queryhelp_hash
        return query


    def inject_query(self, query):
        """
        Manipulate the query an inject it back.
        This can be done to add custom filters using SQLA.
        :param query: A sqlalchemy.orm.Query instance
        """
        from sqlalchemy.orm import Query
        if not isinstance(query, Query):
            raise InputValidationError(
                "{} must be a subclass of {}".format(
                    query, Query
                )
            )
        self._query = query
        self._injected = True

    def distinct(self):
        """
        Asks for distinct rows.
        Does not execute the query!
        If you want a distinct query::

            qb = QueryBuilder(**queryhelp)
            qb.distinct().all() # or
            qb.distinct().get_results_dict()

        :returns: self
        """
        self._query = self.get_query().distinct()
        return self


    def _yield_per(self, batch_size):
        """
        :param count: Number of rows to yield per step

        Yields *count* rows at a time

        :returns: a generator
        """
        return self.get_query().yield_per(batch_size)

    def _all(self):
        return self.get_query().all()

    def _first(self):
        """
        Executes query in the backend asking for one instance.

        :returns: One row of aiida results
        """
        return self.get_query().first()

    def first(self):
        """
        Executes query asking for one instance.
        Use as follows::

            qb = QueryBuilder(**queryhelp)
            qb.first()

        :returns:
            One row of results as a list, order as given by
            order of vertices in  path and projections for vertice
        """
        resultrow = self._first()
        try:
            returnval = [
                    self._get_aiida_res(self._attrkeys_as_in_sql_result[colindex], rowitem)
                    for colindex, rowitem
                    in enumerate(resultrow)
                ]
        except TypeError:
            if resultrow is None:
                returnval = None
            elif len(self._attrkeys_as_in_sql_result) > 1:
                raise Exception(
                    "I have not received an iterable\n"
                    "but the number of projections is > 1"
                )
            # It still returns a list!
            else:
                returnval = [self._get_aiida_res(self._attrkeys_as_in_sql_result[0], resultrow)]
        return returnval



    def count(self):
        """
        Counts the number of rows returned by the backend.

        :returns: the number of rows as an integer
        """
        que = self.get_query()
        return que.count()

    def iterall(self, batch_size=100):
        """
        Same as :func:`QueryBuilderBase.all`, but returns a generator.
        Be aware that this is only safe if no commit will take place during this
        transaction. You might also want to read the SQLAlchemy documentation on
        http://docs.sqlalchemy.org/en/latest/orm/query.html#sqlalchemy.orm.query.Query.yield_per


        :param int batch_size: 
            The size of the batches to ask the backend to batch results in subcollections.
            You can optimize the speed of the query by tuning this parameter.

        :returns: a generator of lists
        """

        if batch_size is not None:
            results = self._yield_per(batch_size)
        else:
            results = self._all()
        try:
            for resultrow in results:
                yield [
                    self._get_aiida_res(self._attrkeys_as_in_sql_result[colindex], rowitem)
                    for colindex, rowitem
                    in enumerate(resultrow)
                ]
        except TypeError:
            # resultrow not an iterable:
            # Checked, result that raises exception is included
            if len(self._attrkeys_as_in_sql_result) > 1:
                raise Exception(
                    "I have not received an iterable\n"
                    "but the number of projections is > 1"
                )
            for rowitem in results:
                yield [self._get_aiida_res(self._attrkeys_as_in_sql_result[0], rowitem)]

    def all(self, batch_size=None):
        """
        Executes the full query with the order of the rows as returned by the backend.
        the order inside each row is given by the order of the vertices in the path
        and the order of the projections for each vertice in the path.

        :param int batch_size: 
            The size of the batches to ask the backend to batch results in subcollections.
            You can optimize the speed of the query by tuning this parameter.
            Leave the default (*None*) if speed is not critical or if you don't know
            what you're doing!

        :returns: a list of lists of all projected entities.
        """

        return list(self.iterall(batch_size=batch_size))


    def dict(self, batch_size=None):
        """
        Executes the full query with the order of the rows as returned by the backend.
        the order inside each row is given by the order of the vertices in the path
        and the order of the projections for each vertice in the path.

        :param int batch_size: 
            The size of the batches to ask the backend to batch results in subcollections.
            You can optimize the speed of the query by tuning this parameter.
            Leave the default (*None*) if speed is not critical or if you don't know
            what you're doing!

        :returns: 
            a list of dictionaries of all projected entities.
            Each dictionary consists of key value pairs, where the key is the tag
            of the vertice and the value a dictionary of key-value pairs where key
            is the entity description (a column name or attribute path)
            and the value the value in the DB.
        
        Usage::

            qb = QueryBuilder()
            qb.append(
                StructureData,
                tag='structure',
                filters={'uuid':{'==':myuuid}},
            )
            qb.append(
                Node,
                descendant_of='structure',
                project=['type', 'id'],  # returns type (string) and id (string)
                tag='descendant'
            )
            
            # Return the dictionaries:
            print "qb.iterdict()"
            for d in qb.iterdict():
                print '>>>', d

        results in the following output::

            qb.iterdict()
            >>> {'descendant': {
                    'type': u'calculation.job.quantumespresso.pw.PwCalculation.',
                    'id': 7716}
                }
            >>> {'descendant': {
                    'type': u'data.remote.RemoteData.',
                    'id': 8510}
                }

        """
        return list(self.iterdict(batch_size=batch_size))


    def iterdict(self, batch_size=100):
        """
        Same as :func:`QueryBuilderBase.dict`, but returns a generator.
        Be aware that this is only safe if no commit will take place during this
        transaction. You might also want to read the SQLAlchemy documentation on
        http://docs.sqlalchemy.org/en/latest/orm/query.html#sqlalchemy.orm.query.Query.yield_per


        :param int batch_size: 
            The size of the batches to ask the backend to batch results in subcollections.
            You can optimize the speed of the query by tuning this parameter.

        :returns: a generator of dictionaries
        """

        if batch_size is not None:
            results = self._yield_per(batch_size=batch_size)
        else:
            results = self._all()
        try:
            for this_result in results:
                yield {
                    tag:{
                        attrkey:self._get_aiida_res(
                                attrkey, this_result[index_in_sql_result]
                            )
                        for attrkey, index_in_sql_result
                        in projected_entities_dict.items()
                    }
                    for tag, projected_entities_dict
                    in self.tag_to_projected_entity_dict.items()
                }
        except TypeError:
            # resultrow not an iterable:
            # Checked, result that raises exception is included
            if len(self._attrkeys_as_in_sql_result) > 1:
                raise Exception(
                    "I have not received an iterable\n"
                    "but the number of projections is > 1"
                )

            for this_result in results:
                yield {
                    tag:{
                        attrkey : self._get_aiida_res(attrkey, this_result)
                        for attrkey, position in projected_entities_dict.items()
                    }
                    for tag, projected_entities_dict in self.tag_to_projected_entity_dict.items()
                }

    def get_results_dict(self):
        """
        Deprecated, use :func:`QueryBuilderBase.dict` or
        :func:`QueryBuilderBase.iterdict` instead
        """
        warnings.warn(
                "get_results_dict will be deprecated in the future"
                "User iterdict for generator or dict for list",
                DeprecationWarning
            )
        
        return self.iterdict()

    @abstractmethod
    def _get_aiida_res(self, key, res):
        """
        Some instance returned by ORM (django or SA) need to be converted
        to Aiida instances (eg nodes)

        :param key: the key that this entry would be returned with
        :param res: the result returned by the query

        :returns: an aiida-compatible instance
        """
        pass

    def inputs(self, **kwargs):
        """
        Join to inputs of previous vertice in path.

        :returns: self
        """
        join_to = self._path[-1]['tag']
        cls = kwargs.pop('cls', self.AiidaNode)
        self.append(cls=cls, input_of=join_to, autotag=True, **kwargs)
        return self

    def outputs(self, **kwargs):
        """
        Join to outputs of previous vertice in path.

        :returns: self
        """
        join_to = self._path[-1]['tag']
        cls = kwargs.pop('cls', self.AiidaNode)
        self.append(cls=cls, output_of=join_to, autotag=True, **kwargs)
        return self

    def children(self, **kwargs):
        """
        Join to children/descendants of previous vertice in path.

        :returns: self
        """
        join_to = self._path[-1]['tag']
        cls = kwargs.pop('cls', self.AiidaNode)
        self.append(cls=cls, descendant_of=join_to, autotag=True, **kwargs)
        return self

    def parents(self, **kwargs):
        """
        Join to parents/ancestors of previous vertice in path.

        :returns: self
        """
        join_to = self._path[-1]['tag']
        cls = kwargs.pop('cls', self.AiidaNode)
        self.append(cls=cls, ancestor_of=join_to, autotag=True, **kwargs)
        return self





# -*- coding: utf-8 -*-
"""
Tests for the export and import routines.
"""
import os
import shutil
import tempfile

from aiida.backends.djsite.db.testbase import AiidaTestCase
from aiida.common.folders import SandboxFolder
from aiida.orm import DataFactory
from aiida.orm import load_node
from aiida.orm.calculation.job import JobCalculation
from aiida.orm.importexport import export, import_data

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class TestPort(AiidaTestCase):
    def test_1(self):
        from aiida.orm import delete_computer

        # Creating a folder for the import/export files
        temp_folder = tempfile.mkdtemp()
        try:
            StructureData = DataFactory('structure')
            sd = StructureData()
            sd.store()

            calc = JobCalculation()
            calc.set_computer(self.computer)
            calc.set_resources({"num_machines": 1, "num_mpiprocs_per_machine": 1})
            calc.store()

            calc.add_link_from(sd)

            pks = [sd.pk, calc.pk]

            attrs = {}
            for pk in pks:
                node = load_node(pk)
                attrs[node.uuid] = dict()
                for k in node.attrs():
                    attrs[node.uuid][k] = node.get_attr(k)

            filename = os.path.join(temp_folder, "export.tar.gz")
            export([calc.dbnode], outfile=filename, silent=True)

            self.tearDownClass()
            self.setUpClass()
            delete_computer(self.computer)

            # NOTE: it is better to load new nodes by uuid, rather than assuming
            # that they will have the first 3 pks. In fact, a recommended policy in
            # databases is that pk always increment, even if you've deleted elements
            import_data(filename, silent=True)
            for uuid in attrs.keys():
                node = load_node(uuid)
                for k in node.attrs():
                    self.assertEquals(attrs[uuid][k], node.get_attr(k))
        finally:
            # Deleting the created temporary folder
            shutil.rmtree(temp_folder, ignore_errors=True)

    def test_2(self):
        """
        Test the check for the export format version.
        """
        import json
        import tarfile

        # Creating a folder for the import/export files
        export_file_tmp_folder = tempfile.mkdtemp()
        unpack_tmp_folder = tempfile.mkdtemp()
        try:
            StructureData = DataFactory('structure')
            sd = StructureData()
            sd.store()

            filename = os.path.join(export_file_tmp_folder, "export.tar.gz")
            export([sd.dbnode], outfile=filename, silent=True)

            with tarfile.open(filename, "r:gz", format=tarfile.PAX_FORMAT) as tar:
                tar.extractall(unpack_tmp_folder)

            with open(os.path.join(unpack_tmp_folder,
                                   'metadata.json'), 'r') as f:
                metadata = json.load(f)
            metadata['export_version'] = 0.0
            with open(os.path.join(unpack_tmp_folder,
                                   'metadata.json'), 'w') as f:
                json.dump(metadata, f)

            with tarfile.open(filename, "w:gz", format=tarfile.PAX_FORMAT) as tar:
                tar.add(unpack_tmp_folder, arcname="")

            self.tearDownClass()
            self.setUpClass()

            with self.assertRaises(ValueError):
                import_data(filename, silent=True)
        finally:
            # Deleting the created temporary folders
            shutil.rmtree(export_file_tmp_folder, ignore_errors=True)
            shutil.rmtree(unpack_tmp_folder, ignore_errors=True)

    def test_3(self):
        """
        Test importing of nodes, that have links to unknown nodes.
        """
        import json
        import tarfile

        # Creating a folder for the import/export files
        temp_folder = tempfile.mkdtemp()
        try:
            StructureData = DataFactory('structure')
            sd = StructureData()
            sd.store()

            filename = os.path.join(temp_folder, "export.tar.gz")
            export([sd.dbnode], outfile=filename, silent=True)

            unpack = SandboxFolder()
            with tarfile.open(filename, "r:gz", format=tarfile.PAX_FORMAT) as tar:
                tar.extractall(unpack.abspath)

            with open(unpack.get_abs_path('data.json'), 'r') as f:
                metadata = json.load(f)
            metadata['links_uuid'].append({
                'output': sd.uuid,
                'input': 'non-existing-uuid',
                'label': 'parent'
            })
            with open(unpack.get_abs_path('data.json'), 'w') as f:
                json.dump(metadata, f)

            with tarfile.open(filename, "w:gz", format=tarfile.PAX_FORMAT) as tar:
                tar.add(unpack.abspath, arcname="")

            self.tearDownClass()
            self.setUpClass()

            with self.assertRaises(ValueError):
                import_data(filename, silent=True)

            import_data(filename, ignore_unknown_nodes=True, silent=True)
        finally:
            # Deleting the created temporary folder
            shutil.rmtree(temp_folder, ignore_errors=True)

    def test_4(self):
        """
        Test control of licenses.
        """
        from aiida.common.exceptions import LicensingException
        from aiida.common.folders import SandboxFolder
        from aiida.orm.importexport import export_tree

        StructureData = DataFactory('structure')
        sd = StructureData()
        sd.source = {'license': 'GPL'}
        sd.store()

        folder = SandboxFolder()
        export_tree([sd.dbnode], folder=folder, silent=True,
                    allowed_licenses=['GPL'])
        # Folder should contain two files of metadata + nodes/
        self.assertEquals(len(folder.get_content_list()), 3)

        folder = SandboxFolder()
        export_tree([sd.dbnode], folder=folder, silent=True,
                    forbidden_licenses=['Academic'])
        # Folder should contain two files of metadata + nodes/
        self.assertEquals(len(folder.get_content_list()), 3)

        folder = SandboxFolder()
        with self.assertRaises(LicensingException):
            export_tree([sd.dbnode], folder=folder, silent=True,
                        allowed_licenses=['CC0'])

        folder = SandboxFolder()
        with self.assertRaises(LicensingException):
            export_tree([sd.dbnode], folder=folder, silent=True,
                        forbidden_licenses=['GPL'])

        def cc_filter(license):
            return license.startswith('CC')

        def gpl_filter(license):
            return license == 'GPL'

        def crashing_filter(license):
            raise NotImplementedError("not implemented yet")

        folder = SandboxFolder()
        with self.assertRaises(LicensingException):
            export_tree([sd.dbnode], folder=folder, silent=True,
                        allowed_licenses=cc_filter)

        folder = SandboxFolder()
        with self.assertRaises(LicensingException):
            export_tree([sd.dbnode], folder=folder, silent=True,
                        forbidden_licenses=gpl_filter)

        folder = SandboxFolder()
        with self.assertRaises(LicensingException):
            export_tree([sd.dbnode], folder=folder, silent=True,
                        allowed_licenses=crashing_filter)

        folder = SandboxFolder()
        with self.assertRaises(LicensingException):
            export_tree([sd.dbnode], folder=folder, silent=True,
                        forbidden_licenses=crashing_filter)



# -*- coding: utf-8 -*-
from aiida.parsers.parser import Parser

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

class BasenwcParser(Parser):

    def parse_with_retrieved(self,retrieved):
        """
        Receives in input a dictionary of retrieved nodes.
        Does all the logic here.
        """
        from aiida.common.exceptions import InvalidOperation
        import os

        output_path = None
        error_path  = None
        try:
            output_path, error_path = self._fetch_output_files(retrieved)
        except InvalidOperation:
            raise
        except IOError as e:
            self.logger.error(e.message)
            return False, ()

        if output_path is None and error_path is None:
            self.logger.error("No output files found")
            return False, ()

        return True, self._get_output_nodes(output_path, error_path)

    def _fetch_output_files(self, retrieved):
        """
        Checks the output folder for standard output and standard error
        files, returns their absolute paths on success.

        :param retrieved: A dictionary of retrieved nodes, as obtained from the
          parser.
        """
        from aiida.common.datastructures import calc_states
        from aiida.common.exceptions import InvalidOperation
        import os

        # check in order not to overwrite anything
#         state = self._calc.get_state()
#         if state != calc_states.PARSING:
#             raise InvalidOperation("Calculation not in {} state"
#                                    .format(calc_states.PARSING) )

        # Check that the retrieved folder is there
        try:
            out_folder = retrieved[self._calc._get_linkname_retrieved()]
        except KeyError:
            raise IOError("No retrieved folder found")

        list_of_files = out_folder.get_folder_list()

        output_path = None
        error_path  = None

        if self._calc._DEFAULT_OUTPUT_FILE in list_of_files:
            output_path = os.path.join( out_folder.get_abs_path('.'),
                                        self._calc._DEFAULT_OUTPUT_FILE )
        if self._calc._DEFAULT_ERROR_FILE in list_of_files:
            error_path  = os.path.join( out_folder.get_abs_path('.'),
                                        self._calc._DEFAULT_ERROR_FILE )

        return output_path, error_path



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import copy

from sqlalchemy import literal
from sqlalchemy.exc import SQLAlchemyError, ProgrammingError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.attributes import flag_modified

from aiida.backends.utils import get_automatic_user
from aiida.backends.sqlalchemy.models.node import DbNode, DbLink, DbPath
from aiida.backends.sqlalchemy.models.comment import DbComment
from aiida.backends.sqlalchemy.models.user import DbUser
from aiida.backends.sqlalchemy.models.computer import DbComputer

from aiida.common.utils import get_new_uuid
from aiida.common.folders import RepositoryFolder
from aiida.common.exceptions import (InternalError, ModificationNotAllowed,
                                     NotExistent, UniquenessError,
                                     ValidationError)
from aiida.common.links import LinkType
from aiida.common.lang import override

from aiida.orm.implementation.general.node import AbstractNode, _NO_DEFAULT
from aiida.orm.implementation.sqlalchemy.computer import Computer
from aiida.orm.implementation.sqlalchemy.group import Group
from aiida.orm.implementation.sqlalchemy.utils import django_filter, get_attr

import aiida.orm.autogroup

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"


class Node(AbstractNode):
    def __init__(self, **kwargs):
        super(Node, self).__init__()

        self._temp_folder = None

        dbnode = kwargs.pop('dbnode', None)

        # Set the internal parameters
        # Can be redefined in the subclasses
        self._init_internal_params()

        if dbnode is not None:
            if not isinstance(dbnode, DbNode):
                raise TypeError("dbnode is not a DbNode instance")
            if dbnode.id is None:
                raise ValueError("If cannot load an aiida.orm.Node instance "
                                 "from an unsaved DbNode object.")
            if kwargs:
                raise ValueError("If you pass a dbnode, you cannot pass any "
                                 "further parameter")

            # If I am loading, I cannot modify it
            self._to_be_stored = False

            self._dbnode = dbnode

            # If this is changed, fix also the importer
            self._repo_folder = RepositoryFolder(section=self._section_name,
                                                 uuid=self._dbnode.uuid)

        else:
            # TODO: allow to get the user from the parameters
            user = get_automatic_user()
            self._dbnode = DbNode(user=user,
                                  uuid=get_new_uuid(),
                                  type=self._plugin_type_string)

            self._to_be_stored = True

            # As creating the temp folder may require some time on slow
            # filesystems, we defer its creation
            self._temp_folder = None
            # Used only before the first save
            self._attrs_cache = {}
            # If this is changed, fix also the importer
            self._repo_folder = RepositoryFolder(section=self._section_name,
                                                 uuid=self.uuid)

            # Automatically set all *other* attributes, if possible, otherwise
            # stop
            self._set_with_defaults(**kwargs)

    @classmethod
    def get_subclass_from_uuid(cls, uuid):
        try:
            node = cls.query(uuid=uuid).one()
        except (NoResultFound, ProgrammingError):
            raise NotExistent("No entry with UUID={} found".format(uuid))
        if not isinstance(node, cls):
            raise NotExistent("UUID={} is not an instance of {}".format(
                uuid, cls.__name__))
        return node

    @classmethod
    def get_subclass_from_pk(cls, pk):
        if not isinstance(pk, int):
            raise ValueError("Incorrect type for int")
        try:
            from aiida.backends.sqlalchemy import session
            node = cls.query(id=pk).one()
        except NoResultFound:
            # DataError is thrown when you pass a string instead of an int for
            # instance
            raise NotExistent("No entry with pk= {} found".format(pk))
        if not isinstance(node, cls):
            raise NotExistent("pk= {} is not an instance of {}".format(
                pk, cls.__name__))
        return node

    def __int__(self):
        if self._to_be_store:
            return None
        else:
            return self._dbnode.id

    @classmethod
    def query(cls, *args, **kwargs):
        q = django_filter(DbNode.aiida_query, **kwargs)
        q = q.reset_joinpoint()

        if cls._plugin_type_string:
            if not cls._plugin_type_string.endswith('.'):
                raise InternalError("The plugin type string does not "
                                    "finish with a dot??")

            # If it is 'calculation.Calculation.', we want to filter
            # for things that start with 'calculation.' and so on
            plug_type = cls._plugin_type_string

            # Remove the implementation.django or sqla part.
            if plug_type.startswith('implementation.'):
                plug_type = '.'.join(plug_type.split('.')[2:])

            pre, sep, _ = plug_type[:-1].rpartition('.')
            superclass_string = "".join([pre, sep])

            q = q.filter(DbNode.type.like("{}%".format(superclass_string)))

        return q

    def _update_db_label_field(self, field_value):
        from aiida.backends.sqlalchemy import session
        self.dbnode.label = field_value
        if not self._to_be_stored:
            self._dbnode.save(commit=False)
            self._increment_version_number_db()

    def _update_db_description_field(self, field_value):
        from aiida.backends.sqlalchemy import session
        self.dbnode.description = field_value
        if not self._to_be_stored:
            self._dbnode.save(commit=False)
            self._increment_version_number_db()

    def _replace_dblink_from(self, src, label, link_type):
        from aiida.backends.sqlalchemy import session
        try:
            self._add_dblink_from(src, label)
        except UniquenessError:
            # I have to replace the link; I do it within a transaction
            self._remove_dblink_from(label)
            self._add_dblink_from(src, label, link_type)

    def _remove_dblink_from(self, label):
        from aiida.backends.sqlalchemy import session
        link = self.dbnode.outputs.filter_by(label=label).first()
        session.delete(link)
        session.commit()

    def _add_dblink_from(self, src, label=None, link_type=LinkType.UNSPECIFIED):
        from aiida.backends.sqlalchemy import session
        if not isinstance(src, Node):
            raise ValueError("src must be a Node instance")
        if self.uuid == src.uuid:
            raise ValueError("Cannot link to itself")

        if self._to_be_stored:
            raise ModificationNotAllowed(
                "Cannot call the internal _add_dblink_from if the "
                "destination node is not stored")
        if src._to_be_stored:
            raise ModificationNotAllowed(
                "Cannot call the internal _add_dblink_from if the "
                "source node is not stored")

        # Check for cycles. This works if the transitive closure is enabled; if
        # it isn't, this test will never fail, but then having a circular link
        # is not meaningful but does not pose a huge threat
        #
        # I am linking src->self; a loop would be created if a DbPath exists
        # already in the TC table from self to src
        c = session.query(literal(True)).filter(DbPath.query
                                                .filter_by(parent_id=self.dbnode.id, child_id=src.dbnode.id)
                                                .exists()).scalar()
        if c:
            raise ValueError(
                "The link you are attempting to create would generate a loop")

        if label is None:
            autolabel_idx = 1

            existing_from_autolabels = session.query(DbLink.label).filter(
                DbLink.output_id == self.dbnode.id,
                DbLink.label.like("link%")
            )

            while "link_{}".format(autolabel_idx) in existing_from_autolabels:
                autolabel_idx += 1

            safety_counter = 0
            while True:
                safety_counter += 1
                if safety_counter > 3:
                    # Well, if you have more than 100 concurrent addings
                    # to the same node, you are clearly doing something wrong...
                    raise InternalError("Hey! We found more than 100 concurrent"
                                        " adds of links "
                                        "to the same nodes! Are you really doing that??")
                try:
                    self._do_create_link(src, "link_{}".format(autolabel_idx), link_type)
                    break
                except UniquenessError:
                    # Retry loop until you find a new loop
                    autolabel_idx += 1
        else:
            self._do_create_link(src, label, link_type)

    def _do_create_link(self, src, label, link_type):
        from aiida.backends.sqlalchemy import session
        try:
            link = DbLink(input_id=src.dbnode.id, output_id=self.dbnode.id,
                          label=label, type=link_type.value)
            session.add(link)
            session.commit()
        except SQLAlchemyError as e:
            session.rollback()
            raise UniquenessError("There is already a link with the same "
                                  "name (raw message was {})"
                                  "".format(e))

    def get_inputs(self, node_type=None, also_labels=False, only_in_db=False,
                   link_type=None):

        link_filter = {'output': self.dbnode}
        if link_type is not None:
            link_filter['type'] = link_type.value
        inputs_list = [(i.label, i.input.get_aiida_class()) for i in
                       DbLink.query.filter_by(output=self.dbnode)
                           .distinct().all()]

        if not only_in_db:
            # Needed for the check
            input_list_keys = [i[0] for i in inputs_list]

            for k, v in self._inputlinks_cache.iteritems():
                src = v[0]
                if k in input_list_keys:
                    raise InternalError("There exist a link with the same name "
                                        "'{}' both in the DB and in the internal "
                                        "cache for node pk= {}!".format(k, self.id))
                inputs_list.append((k, src))

        if node_type is None:
            filtered_list = inputs_list
        else:
            filtered_list = [i for i in inputs_list if isinstance(i[1], node_type)]

        if also_labels:
            return list(filtered_list)
        else:
            return [i[1] for i in filtered_list]

    @override
    def get_outputs(self, type=None, also_labels=False, link_type=None):

        link_filter = {'input': self.dbnode}
        if link_type is not None:
            link_filter['type'] = link_type.value
        outputs_list = ((i.label, i.output.get_aiida_class()) for i in
                        DbLink.query.filter_by(**link_filter).distinct().all())

        if type is None:
            if also_labels:
                return list(outputs_list)
            else:
                return [i[1] for i in outputs_list]
        else:
            filtered_list = (i for i in outputs_list if isinstance(i[1], type))
            if also_labels:
                return list(filtered_list)
            else:
                return [i[1] for i in filtered_list]

    def set_computer(self, computer):
        if self._to_be_stored:
            computer = DbComputer.get_dbcomputer(computer)
            self.dbnode.dbcomputer = computer
        else:
            raise ModificationNotAllowed(
                "Node with uuid={} was already stored".format(self.uuid))

    def _set_attr(self, key, value):
        if self._to_be_stored:
            self._attrs_cache[key] = copy.deepcopy(value)
        else:
            self.dbnode.set_attr(key, value)
            self._increment_version_number_db()

    def _del_attr(self, key):
        if self._to_be_stored:
            try:
                del self._attrs_cache[key]
            except KeyError:
                raise AttributeError(
                    "Attribute {} does not exist".format(key))
        else:
            self.dbnode.del_attr(key)
            self._increment_version_number_db()

    def get_attr(self, key, default=_NO_DEFAULT):
        exception = AttributeError("Attribute '{}' does not exist".format(key))

        has_default = default is not _NO_DEFAULT
        if self._to_be_stored:
            try:
                return self._attrs_cache[key]
            except KeyError:
                if has_default:
                    return default
                raise exception
        else:
            try:
                return get_attr(self.dbnode.attributes, key)
            except (KeyError, IndexError):
                if has_default:
                    return default
                else:
                    raise exception

    def set_extra(self, key, value, exclusive=False):
        # TODO SP: validate key
        # TODO SP: handle exclusive (what to do in case the key already exist
        # ?)
        if self._to_be_stored:
            raise ModificationNotAllowed(
                "The extras of a node can be set only after "
                "storing the node")

        self.dbnode.set_extra(key, value)
        self._increment_version_number_db()

    def get_extra(self, key, default=None):
        # TODO SP: in the Django implementation, if the node is not stored,
        # we can't get an extra. In the SQLA one, because this is simply a
        # column, we could still return one if it exists.
        try:
            return get_attr(self.dbnode.extras, key)
        except (KeyError, IndexError) as e:
            if default:
                return default
            else:
                raise AttributeError

    def get_extras(self):
        return self.dbnode.extras

    def del_extra(self, key):
        self.dbnode.del_extra(key)

    def extras(self):
        return self.dbnode.extras

    def iterextras(self):
        return self.dbnode.extras.iteritems()

    def iterattrs(self):
        # TODO: check what happens if someone stores the object while
        #        the iterator is being used!
        if self._to_be_stored:
            it_items = self._attrs_cache.iteritems()
        else:
            it_items = self.dbnode.attributes.iteritems()

        for k, v in it_items:
            yield (k, v)

    def get_attrs(self):
        return self.dbnode.attributes

    def attrs(self):
        if self._to_be_stored:
            it = self._attrs_cache.iterkeys()
        else:
            it = self.dbnode.attributes.iterkeys()
        for k in it:
            yield k

    def add_comment(self, content, user=None):
        from aiida.backends.sqlalchemy import session
        if self._to_be_stored:
            raise ModificationNotAllowed("Comments can be added only after "
                                         "storing the node")

        comment = DbComment(dbnode=self._dbnode, user=user, content=content)
        session.add(comment)
        session.commit()

    def get_comment_obj(self, id=None, user=None):
        dbcomments_query = DbComment.query.filter_by(dbnode=self._dbnode)

        if id is not None:
            dbcomments_query = dbcomments_query.filter_by(id=id)
        if user is not None:
            dbcomments_query = dbcomments_query.filter_by(user=user)

        dbcomments = dbcomments_query.all()
        comments = []
        from aiida.orm.implementation.sqlalchemy.comment import Comment
        for dbcomment in dbcomments:
            comments.append(Comment(dbcomment=dbcomment))
        return comments

    def get_comments(self, pk=None):
        comments = self._get_dbcomments(pk)

        return [{
            "pk": c.id,
            "user__email": c.user.email,
            "ctime": c.ctime,
            "mtime": c.mtime,
            "content": c.content
        } for c in comments ]

    def _get_dbcomments(self, pk=None, with_user=False):
        comments = DbComment.query.filter_by(dbnode=self._dbnode)

        if pk is not None:
            try:
                correct = all([isinstance(_, int) for _ in pk])
                if not correct:
                    raise ValueError('id must be an integer or a list of integers')
                comments = comments.filter(DbComment.id.in_(pk))
            except TypeError:
                if not isinstance(pk, int):
                    raise ValueError('id must be an integer or a list of integers')

                comments = comments.filter_by(id=pk)

        if with_user:
            comments.join(DbUser)

        comments = comments.order_by('id').all()

        return comments

    def _update_comment(self, new_field, comment_pk, user):
        comment = DbComment.query.filter_by(dbnode=self._dbnode,
                                            id=comment_pk, user=user).first()

        if not isinstance(new_field, basestring):
            raise ValueError("Non string comments are not accepted")

        if not comment:
            raise NotExistent("Found no comment for user {} and id {}".format(
                user, comment_pk))

        comment.content = new_field
        comment.save()

    def _remove_comment(self, comment_pk, user):
        comment = DbComment.query.filter_by(dbnode=self._dbnode, id=comment_pk).first()
        if comment:
            comment.delete()

    def _increment_version_number_db(self):
        self._dbnode.nodeversion = DbNode.nodeversion + 1
        self._dbnode.save()

    def copy(self):
        newobject = self.__class__()
        newobject.dbnode.type = self.dbnode.type  # Inherit type
        newobject.dbnode.label = self.dbnode.label  # Inherit label
        # TODO: add to the description the fact that this was a copy?
        newobject.dbnode.description = self.dbnode.description  # Inherit description
        newobject.dbnode.dbcomputer = self.dbnode.dbcomputer  # Inherit computer

        for k, v in self.iterattrs():
            newobject._set_attr(k, v)

        for path in self.get_folder_list():
            newobject.add_path(self.get_abs_path(path), path)

        return newobject

    @property
    def pk(self):
        return self.dbnode.id

    @property
    def id(self):
        return self.dbnode.id

    @property
    def dbnode(self):
        return self._dbnode

    def store_all(self, with_transaction=True):
        """
        Store the node, together with all input links, if cached, and also the
        linked nodes, if they were not stored yet.

        :parameter with_transaction: if False, no transaction is used. This
          is meant to be used ONLY if the outer calling function has already
          a transaction open!
        """

        if not self._to_be_stored:
            raise ModificationNotAllowed(
                "Node with pk= {} was already stored".format(self.id))

        # For each parent, check that all its inputs are stored
        for link in self._inputlinks_cache:
            try:
                parent_node = self._inputlinks_cache[link][0]
                parent_node._check_are_parents_stored()
            except ModificationNotAllowed:
                raise ModificationNotAllowed("Parent node (UUID={}) has "
                                             "unstored parents, cannot proceed (only direct parents "
                                             "can be unstored and will be stored by store_all, not "
                                             "grandparents or other ancestors".format(parent_node.uuid))

        self._store_input_nodes()
        self.store(with_transaction=False)
        self._store_cached_input_links(with_transaction=False)
        from aiida.backends.sqlalchemy import session

        if with_transaction:
            try:
                session.commit()
            except SQLAlchemyError as e:
                session.rollback()

        return self

    def _store_input_nodes(self):
        """
        Find all input nodes, and store them, checking that they do not
        have unstored inputs in turn.

        :note: this function stores all nodes without transactions; always
          call it from within a transaction!
        """
        if not self._to_be_stored:
            raise ModificationNotAllowed(
                "_store_input_nodes can be called only if the node is "
                "unstored (node {} is stored, instead)".format(self.id))

        for link in self._inputlinks_cache:
            parent = self._inputlinks_cache[link][0]
            if not parent.is_stored:
                parent.store(with_transaction=False)

    def _check_are_parents_stored(self):
        """
        Check if all parents are already stored, otherwise raise.

        :raise ModificationNotAllowed: if one of the input nodes in not already
          stored.
        """
        # Preliminary check to verify that inputs are stored already
        for link in self._inputlinks_cache:
            if not self._inputlinks_cache[link][0].is_stored:
                raise ModificationNotAllowed(
                    "Cannot store the input link '{}' because the "
                    "source node is not stored. Either store it first, "
                    "or call _store_input_links with the store_parents "
                    "parameter set to True".format(link))

    def _store_cached_input_links(self, with_transaction=True):
        """
        Store all input links that are in the local cache, transferring them
        to the DB.

        :note: This can be called only if all parents are already stored.

        :note: Links are stored only after the input nodes are stored. Moreover,
            link storage is done in a transaction, and if one of the links
            cannot be stored, an exception is raised and *all* links will remain
            in the cache.

        :note: This function can be called only after the node is stored.
           After that, it can be called multiple times, and nothing will be
           executed if no links are still in the cache.

        :parameter with_transaction: if False, no transaction is used. This
          is meant to be used ONLY if the outer calling function has already
          a transaction open!
        """

        if self._to_be_stored:
            raise ModificationNotAllowed(
                "Node with pk= {} is not stored yet".format(self.id))

        # This raises if there is an unstored node.
        self._check_are_parents_stored()
        # I have to store only those links where the source is already
        # stored
        links_to_store = list(self._inputlinks_cache.keys())

        for label in links_to_store:
            src, link_type = self._inputlinks_cache[label]
            self._add_dblink_from(src, label, link_type)
        # If everything went smoothly, clear the entries from the cache.
        # I do it here because I delete them all at once if no error
        # occurred; otherwise, links will not be stored and I
        # should not delete them from the cache (but then an exception
        # would have been raised, and the following lines are not executed)
        self._inputlinks_cache.clear()

        from aiida.backends.sqlalchemy import session
        if with_transaction:
            try:
                session.commit()
            except SQLAlchemyError as e:
                session.rollback()

    def store(self, with_transaction=True):
        """
        Store a new node in the DB, also saving its repository directory
        and attributes.

        After being called attributes cannot be
        changed anymore! Instead, extras can be changed only AFTER calling
        this store() function.

        :note: After successful storage, those links that are in the cache, and
            for which also the parent node is already stored, will be
            automatically stored. The others will remain unstored.

        :parameter with_transaction: if False, no transaction is used. This
          is meant to be used ONLY if the outer calling function has already
          a transaction open!
        """
        # TODO: This needs to be generalized, allowing for flexible methods
        # for storing data and its attributes.
        if self._to_be_stored:
            self._validate()

            self._check_are_parents_stored()

            # I save the corresponding django entry
            # I set the folder
            # NOTE: I first store the files, then only if this is successful,
            # I store the DB entry. In this way,
            # I assume that if a node exists in the DB, its folder is in place.
            # On the other hand, periodically the user might need to run some
            # bookkeeping utility to check for lone folders.
            self._repository_folder.replace_with_folder(
                self._get_temp_folder().abspath, move=True, overwrite=True)

            try:
                self._dbnode.save(commit=False)
                # Save its attributes 'manually' without incrementing
                # the version for each add.
                self.dbnode.attributes = self._attrs_cache
                flag_modified(self.dbnode, "attributes")
                # This should not be used anymore: I delete it to
                # possibly free memory
                del self._attrs_cache

                self._temp_folder = None
                self._to_be_stored = False

                # Here, I store those links that were in the cache and
                # that are between stored nodes.
                self._store_cached_input_links(with_transaction=False)

                if with_transaction:
                    self.dbnode.session.commit()

            # This is one of the few cases where it is ok to do a 'global'
            # except, also because I am re-raising the exception
            except:
                # I put back the files in the sandbox folder since the
                # transaction did not succeed
                self._get_temp_folder().replace_with_folder(
                    self._repository_folder.abspath, move=True, overwrite=True)
                raise

            # Set up autogrouping used be verdi run
            autogroup = aiida.orm.autogroup.current_autogroup
            grouptype = aiida.orm.autogroup.VERDIAUTOGROUP_TYPE

            if autogroup is not None:
                if not isinstance(autogroup, aiida.orm.autogroup.Autogroup):
                    raise ValidationError("current_autogroup is not an AiiDA Autogroup")

                if autogroup.is_to_be_grouped(self):
                    group_name = autogroup.get_group_name()
                    if group_name is not None:
                        g = Group.get_or_create(name=group_name, type_string=grouptype)[0]
                        g.add_nodes(self)

        return self

    @property
    def has_children(self):
        #~ from aiida.backends.sqlalchemy.model.node import DbNode
        return self.dbnode.outputs_q.all()
        #~ return session.query(literal(True)).filter(
            #~ self.dbnode.child_paths.exists()).scalar() or False

    @property
    def has_parents(self):
        return session.query(literal(True)).filter(
            self.dbnode.parent_paths.exists()).scalar() or False

    @property
    def uuid(self):
        return unicode(self.dbnode.uuid)



# -*- coding: utf-8 -*-
"""
Classes needed for tests.
Must be here because subclasses of 'Node' must be within aiida.orm
"""
from aiida.orm.calculation import Calculation

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class myNodeWithFields(Calculation):
    # State can be updated even after storing
    _updatable_attributes = ('state',)



# -*- coding: utf-8 -*-
"""
This file contains the main routines to submit, check and retrieve calculation
results. These are general and contain only the main logic; where appropriate,
the routines make reference to the suitable plugins for all
plugin-specific operations.
"""
from aiida.common.datastructures import calc_states
from aiida.scheduler.datastructures import job_states
from aiida.common.exceptions import (
    AuthenticationError,
    ConfigurationError,
    ModificationNotAllowed,
)
from aiida.common import aiidalogger
from aiida.common.links import LinkType
from aiida.orm import load_node


__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

execlogger = aiidalogger.getChild('execmanager')


def update_running_calcs_status(authinfo):
    """
    Update the states of calculations in WITHSCHEDULER status belonging
    to user and machine as defined in the 'dbauthinfo' table.
    """
    from aiida.orm import JobCalculation, Computer
    from aiida.scheduler.datastructures import JobInfo
    from aiida.utils.logger import get_dblogger_extra

    if not authinfo.enabled:
        return

    execlogger.debug("Updating running calc status for user {} "
                     "and machine {}".format(
        authinfo.aiidauser.email, authinfo.dbcomputer.name))

    # This returns an iterator over aiida JobCalculation objects
    calcs_to_inquire = list(JobCalculation._get_all_with_state(
        state=calc_states.WITHSCHEDULER,
        computer=authinfo.dbcomputer,
        user=authinfo.aiidauser)
    )

    # NOTE: no further check is done that machine and
    # aiidauser are correct for each calc in calcs
    s = Computer(dbcomputer=authinfo.dbcomputer).get_scheduler()
    t = authinfo.get_transport()

    computed = []

    # I avoid to open an ssh connection if there are
    # no calcs with state WITHSCHEDULER
    if len(calcs_to_inquire):
        jobids_to_inquire = [str(c.get_job_id()) for c in calcs_to_inquire]

        # Open connection
        with t:
            s.set_transport(t)
            # TODO: Check if we are ok with filtering by job (to make this work,
            # I had to remove the check on the retval for getJobs,
            # because if the job has computed and is not in the output of
            # qstat, it gives a nonzero retval)

            # TODO: catch SchedulerError exception and do something
            # sensible (at least, skip this computer but continue with
            # following ones, and set a counter; set calculations to
            # UNKNOWN after a while?
            if s.get_feature('can_query_by_user'):
                found_jobs = s.getJobs(user="$USER", as_dict=True)
            else:
                found_jobs = s.getJobs(jobs=jobids_to_inquire, as_dict=True)

            # I update the status of jobs

            for c in calcs_to_inquire:
                try:
                    logger_extra = get_dblogger_extra(c)
                    t._set_logger_extra(logger_extra)

                    jobid = c.get_job_id()
                    if jobid is None:
                        execlogger.error("JobCalculation {} is WITHSCHEDULER "
                                         "but no job id was found!".format(
                            c.pk), extra=logger_extra)
                        continue

                    # I check if the calculation to be checked (c)
                    # is in the output of qstat
                    if jobid in found_jobs:
                        # jobinfo: the information returned by
                        # qstat for this job
                        jobinfo = found_jobs[jobid]
                        execlogger.debug("Inquirying calculation {} (jobid "
                                         "{}): it has job_state={}".format(
                            c.pk, jobid, jobinfo.job_state), extra=logger_extra)
                        # For the moment, FAILED is not defined
                        if jobinfo.job_state in [job_states.DONE]:  # , job_states.FAILED]:
                            computed.append(c)
                            try:
                                c._set_state(calc_states.COMPUTED)
                            except ModificationNotAllowed:
                                # Someone already set it, just skip
                                pass

                        ## Do not set the WITHSCHEDULER state multiple times,
                        ## this would raise a ModificationNotAllowed
                        # else:
                        # c._set_state(calc_states.WITHSCHEDULER)

                        c._set_scheduler_state(jobinfo.job_state)

                        c._set_last_jobinfo(jobinfo)
                    else:
                        execlogger.debug("Inquirying calculation {} (jobid "
                                         "{}): not found, assuming "
                                         "job_state={}".format(
                            c.pk, jobid, job_states.DONE), extra=logger_extra)

                        # calculation c is not found in the output of qstat
                        computed.append(c)
                        c._set_scheduler_state(job_states.DONE)
                except Exception as e:
                    # TODO: implement a counter, after N retrials
                    # set it to a status that
                    # requires the user intervention
                    execlogger.warning(
                        "There was an exception for "
                        "calculation {} ({}): {}".format(
                            c.pk, e.__class__.__name__, e.message
                        ), extra=logger_extra)
                    continue

            for c in computed:
                try:
                    logger_extra = get_dblogger_extra(c)
                    try:
                        detailed_jobinfo = s.get_detailed_jobinfo(
                            jobid=c.get_job_id())
                    except NotImplementedError:
                        detailed_jobinfo = (
                            u"AiiDA MESSAGE: This scheduler does not implement "
                            u"the routine get_detailed_jobinfo to retrieve "
                            u"the information on "
                            u"a job after it has finished.")
                    last_jobinfo = c._get_last_jobinfo()
                    if last_jobinfo is None:
                        last_jobinfo = JobInfo()
                        last_jobinfo.job_id = c.get_job_id()
                        last_jobinfo.job_state = job_states.DONE
                    last_jobinfo.detailedJobinfo = detailed_jobinfo
                    c._set_last_jobinfo(last_jobinfo)
                except Exception as e:
                    execlogger.warning("There was an exception while "
                                       "retrieving the detailed jobinfo "
                                       "for calculation {} ({}): {}".format(
                        c.pk, e.__class__.__name__, e.message),
                                       extra=logger_extra)
                    continue
                finally:
                    # Set the state to COMPUTED as the very last thing
                    # of this routine; no further change should be done after
                    # this, so that in general the retriever can just
                    # poll for this state, if we want to.
                    try:
                        c._set_state(calc_states.COMPUTED)
                    except ModificationNotAllowed:
                        # Someone already set it, just skip
                        pass

    return computed


def retrieve_jobs():
    from aiida.orm import JobCalculation, Computer
    from aiida.backends.utils import get_authinfo

    # I create a unique set of pairs (computer, aiidauser)
    computers_users_to_check = list(
        JobCalculation._get_all_with_state(
            state=calc_states.COMPUTED,
            only_computer_user_pairs=True,
            only_enabled=True)
    )

    for computer, aiidauser in computers_users_to_check:
        execlogger.debug("({},{}) pair to check".format(
            aiidauser.email, computer.name))
        try:
            authinfo = get_authinfo(computer.dbcomputer, aiidauser._dbuser)
            retrieve_computed_for_authinfo(authinfo)
        except Exception as e:
            msg = ("Error while retrieving calculation status for "
                   "aiidauser={} on computer={}, "
                   "error type is {}, error message: {}".format(
                aiidauser.email,
                computer.name,
                e.__class__.__name__, e.message))
            execlogger.error(msg)
            # Continue with next computer
            continue


# in daemon
def update_jobs():
    """
    calls an update for each set of pairs (machine, aiidauser)
    """
    from aiida.orm import JobCalculation, Computer, User
    from aiida.backends.utils import get_authinfo


    # I create a unique set of pairs (computer, aiidauser)
    computers_users_to_check = list(
        JobCalculation._get_all_with_state(
            state=calc_states.WITHSCHEDULER,
            only_computer_user_pairs=True,
            only_enabled=True
        )
    )

    for computer, aiidauser in computers_users_to_check:

        execlogger.debug("({},{}) pair to check".format(
            aiidauser.email, computer.name))

        try:
            authinfo = get_authinfo(computer.dbcomputer, aiidauser._dbuser)
            computed_calcs = update_running_calcs_status(authinfo)
        except Exception as e:
            msg = ("Error while updating calculation status "
                   "for aiidauser={} on computer={}, "
                   "error type is {}, error message: {}".format(
                aiidauser.email,
                computer.name,
                e.__class__.__name__, e.message))
            execlogger.error(msg)
            # Continue with next computer
            continue


def submit_jobs():
    """
    Submit all jobs in the TOSUBMIT state.
    """
    from aiida.orm import JobCalculation, Computer, User
    from aiida.utils.logger import get_dblogger_extra
    from aiida.backends.utils import get_authinfo

    computers_users_to_check = list(JobCalculation._get_all_with_state(
            state=calc_states.TOSUBMIT,
            only_computer_user_pairs=True,
            only_enabled=True
        )
    )

    for computer, aiidauser in computers_users_to_check:
        #~ user = User.search_for_users(id=dbuser_id)
        #~ computer = Computer.get(dbcomputer_id)
        execlogger.debug("({},{}) pair to submit".format(
            aiidauser.email, computer.name))

        try:
            try:
                authinfo = get_authinfo(computer.dbcomputer, aiidauser._dbuser)
            except AuthenticationError:
                # TODO!!
                # Put each calculation in the SUBMISSIONFAILED state because
                # I do not have AuthInfo to submit them
                calcs_to_inquire = JobCalculation._get_all_with_state(
                    state=calc_states.TOSUBMIT,
                    computer=computer, user=aiidauser)
                for calc in calcs_to_inquire:
                    try:
                        calc._set_state(calc_states.SUBMISSIONFAILED)
                    except ModificationNotAllowed:
                        # Someone already set it, just skip
                        pass
                    logger_extra = get_dblogger_extra(calc)
                    execlogger.error("Submission of calc {} failed, "
                                     "computer pk= {} ({}) is not configured "
                                     "for aiidauser {}".format(
                        calc.pk, computer.pk, computer.get_name(),
                        aiidauser.email),
                                     extra=logger_extra)
                # Go to the next (dbcomputer,aiidauser) pair
                continue

            submitted_calcs = submit_jobs_with_authinfo(authinfo)
        except Exception as e:
            import traceback

            msg = ("Error while submitting jobs "
                   "for aiidauser={} on computer={}, "
                   "error type is {}, traceback: {}".format(
                aiidauser.email,
                computer.name,
                e.__class__.__name__, traceback.format_exc()))
            print msg
            execlogger.error(msg)
            # Continue with next computer
            continue


def submit_jobs_with_authinfo(authinfo):
    """
    Submit jobs in TOSUBMIT status belonging
    to user and machine as defined in the 'dbauthinfo' table.
    """
    from aiida.orm import JobCalculation
    from aiida.utils.logger import get_dblogger_extra

    if not authinfo.enabled:
        return

    execlogger.debug("Submitting jobs for user {} "
                     "and machine {}".format(
        authinfo.aiidauser.email, authinfo.dbcomputer.name))

    # This returns an iterator over aiida JobCalculation objects
    calcs_to_inquire = list(JobCalculation._get_all_with_state(
        state=calc_states.TOSUBMIT,
        computer=authinfo.dbcomputer,
        user=authinfo.aiidauser))

    # I avoid to open an ssh connection if there are
    # no calcs with state WITHSCHEDULER
    if len(calcs_to_inquire):
        # Open connection
        try:
            # I do it here so that the transport is opened only once per computer
            with authinfo.get_transport() as t:
                for c in calcs_to_inquire:
                    logger_extra = get_dblogger_extra(c)
                    t._set_logger_extra(logger_extra)

                    try:
                        submit_calc(calc=c, authinfo=authinfo, transport=t)
                    except Exception as e:
                        # TODO: implement a counter, after N retrials
                        # set it to a status that
                        # requires the user intervention
                        execlogger.warning("There was an exception for "
                                           "calculation {} ({}): {}".format(
                            c.pk, e.__class__.__name__, e.message))
                        # I just proceed to the next calculation
                        continue
        # Catch exceptions also at this level (this happens only if there is
        # a problem opening the transport in the 'with t' statement,
        # because any other exception is caught and skipped above
        except Exception as e:
            import traceback
            from aiida.utils.logger import get_dblogger_extra

            for calc in calcs_to_inquire:
                logger_extra = get_dblogger_extra(calc)
                try:
                    calc._set_state(calc_states.SUBMISSIONFAILED)
                except ModificationNotAllowed:
                    # Someone already set it, just skip
                    pass

                execlogger.error("Submission of calc {} failed, check also the "
                                 "log file! Traceback: {}".format(calc.pk,
                                                                  traceback.format_exc()),
                                 extra=logger_extra)
            raise


def submit_calc(calc, authinfo, transport=None):
    """
    Submit a calculation

    :note: if no transport is passed, a new transport is opened and then
        closed within this function. If you want to use an already opened
        transport, pass it as further parameter. In this case, the transport
        has to be already open, and must coincide with the transport of the
        the computer defined by the authinfo.

    :param calc: the calculation to submit
        (an instance of the aiida.orm.JobCalculation class)
    :param authinfo: the authinfo for this calculation.
    :param transport: if passed, must be an already opened transport. No checks
        are done on the consistency of the given transport with the transport
        of the computer defined in the authinfo.
    """
    from aiida.orm import Code, Computer
    from aiida.common.folders import SandboxFolder
    from aiida.common.exceptions import (
        InputValidationError)
    from aiida.orm.data.remote import RemoteData
    from aiida.utils.logger import get_dblogger_extra

    if not authinfo.enabled:
        return

    logger_extra = get_dblogger_extra(calc)

    if transport is None:
        t = authinfo.get_transport()
        must_open_t = True
    else:
        t = transport
        must_open_t = False

    t._set_logger_extra(logger_extra)

    if calc._has_cached_links():
        raise ValueError("Cannot submit calculation {} because it has "
                         "cached input links! If you "
                         "just want to test the submission, use the "
                         "test_submit() method, otherwise store all links"
                         "first".format(calc.pk))

    # Double check, in the case the calculation was 'killed' (and therefore
    # put in the 'FAILED' state) in the meantime
    # Do it as near as possible to the state change below (it would be
    # even better to do it with some sort of transaction)
    if calc.get_state() != calc_states.TOSUBMIT:
        raise ValueError("Can only submit calculations with state=TOSUBMIT! "
                         "(state of calc {} is {} instead)".format(calc.pk,
                                                                   calc.get_state()))
    # I start to submit the calculation: I set the state
    try:
        calc._set_state(calc_states.SUBMITTING)
    except ModificationNotAllowed:
        raise ValueError("The calculation has already been submitted by "
                         "someone else!")

    try:
        if must_open_t:
            t.open()

        s = Computer(dbcomputer=authinfo.dbcomputer).get_scheduler()
        s.set_transport(t)

        computer = calc.get_computer()

        with SandboxFolder() as folder:
            calcinfo, script_filename = calc._presubmit(
                folder, use_unstored_links=False)

            codes_info = calcinfo.codes_info
            input_codes = [load_node(_.code_uuid, parent_class=Code)
                           for _ in codes_info ]

            for code in input_codes:
                if not code.can_run_on(computer):
                    raise InputValidationError(
                        "The selected code {} for calculation "
                        "{} cannot run on computer {}".
                        format(code.pk, calc.pk, computer.name))

            # After this call, no modifications to the folder should be done
            calc._store_raw_input_folder(folder.abspath)

            # NOTE: some logic is partially replicated in the 'test_submit'
            # method of JobCalculation. If major logic changes are done
            # here, make sure to update also the test_submit routine
            remote_user = t.whoami()
            # TODO Doc: {username} field
            # TODO: if something is changed here, fix also 'verdi computer test'
            remote_working_directory = authinfo.get_workdir().format(
                username=remote_user)
            if not remote_working_directory.strip():
                raise ConfigurationError(
                    "[submission of calc {}] "
                    "No remote_working_directory configured for computer "
                    "'{}'".format(calc.pk, computer.name))

            # If it already exists, no exception is raised
            try:
                t.chdir(remote_working_directory)
            except IOError:
                execlogger.debug(
                    "[submission of calc {}] "
                    "Unable to chdir in {}, trying to create it".
                    format(calc.pk, remote_working_directory),
                    extra=logger_extra)
                try:
                    t.makedirs(remote_working_directory)
                    t.chdir(remote_working_directory)
                except (IOError, OSError) as e:
                    raise ConfigurationError(
                        "[submission of calc {}] "
                        "Unable to create the remote directory {} on "
                        "computer '{}': {}".
                        format(calc.pk, remote_working_directory, computer.name,
                               e.message))
            # Store remotely with sharding (here is where we choose
            # the folder structure of remote jobs; then I store this
            # in the calculation properties using _set_remote_dir
            # and I do not have to know the logic, but I just need to
            # read the absolute path from the calculation properties.
            t.mkdir(calcinfo.uuid[:2], ignore_existing=True)
            t.chdir(calcinfo.uuid[:2])
            t.mkdir(calcinfo.uuid[2:4], ignore_existing=True)
            t.chdir(calcinfo.uuid[2:4])
            t.mkdir(calcinfo.uuid[4:])
            t.chdir(calcinfo.uuid[4:])
            workdir = t.getcwd()
            # I store the workdir of the calculation for later file
            # retrieval
            calc._set_remote_workdir(workdir)

            # I first create the code files, so that the code can put
            # default files to be overwritten by the plugin itself.
            # Still, beware! The code file itself could be overwritten...
            # But I checked for this earlier.
            for code in input_codes:
                if code.is_local():
                    # Note: this will possibly overwrite files
                    for f in code.get_folder_list():
                        t.put(code.get_abs_path(f), f)
                    t.chmod(code.get_local_executable(), 0755)  # rwxr-xr-x

            # copy all files, recursively with folders
            for f in folder.get_content_list():
                execlogger.debug("[submission of calc {}] "
                                 "copying file/folder {}...".format(calc.pk, f),
                                 extra=logger_extra)
                t.put(folder.get_abs_path(f), f)

            # local_copy_list is a list of tuples,
            # each with (src_abs_path, dest_rel_path)
            # NOTE: validation of these lists are done
            # inside calc._presubmit()
            local_copy_list = calcinfo.local_copy_list
            remote_copy_list = calcinfo.remote_copy_list
            remote_symlink_list = calcinfo.remote_symlink_list

            if local_copy_list is not None:
                for src_abs_path, dest_rel_path in local_copy_list:
                    execlogger.debug("[submission of calc {}] "
                                     "copying local file/folder to {}".format(
                        calc.pk, dest_rel_path),
                                     extra=logger_extra)
                    t.put(src_abs_path, dest_rel_path)

            if remote_copy_list is not None:
                for (remote_computer_uuid, remote_abs_path,
                     dest_rel_path) in remote_copy_list:
                    if remote_computer_uuid == computer.uuid:
                        execlogger.debug("[submission of calc {}] "
                                         "copying {} remotely, directly on the machine "
                                         "{}".format(calc.pk, dest_rel_path, computer.name))
                        try:
                            t.copy(remote_abs_path, dest_rel_path)
                        except (IOError, OSError):
                            execlogger.warning("[submission of calc {}] "
                                               "Unable to copy remote resource from {} to {}! "
                                               "Stopping.".format(calc.pk,
                                                                  remote_abs_path, dest_rel_path),
                                               extra=logger_extra)
                            raise
                    else:
                        # TODO: implement copy between two different
                        # machines!
                        raise NotImplementedError(
                            "[presubmission of calc {}] "
                            "Remote copy between two different machines is "
                            "not implemented yet".format(calc.pk))

            if remote_symlink_list is not None:
                for (remote_computer_uuid, remote_abs_path,
                     dest_rel_path) in remote_symlink_list:
                    if remote_computer_uuid == computer.uuid:
                        execlogger.debug("[submission of calc {}] "
                                         "copying {} remotely, directly on the machine "
                                         "{}".format(calc.pk, dest_rel_path, computer.name))
                        try:
                            t.symlink(remote_abs_path, dest_rel_path)
                        except (IOError, OSError):
                            execlogger.warning("[submission of calc {}] "
                                               "Unable to create remote symlink from {} to {}! "
                                               "Stopping.".format(calc.pk,
                                                                  remote_abs_path, dest_rel_path),
                                               extra=logger_extra)
                            raise
                    else:
                        raise IOError("It is not possible to create a symlink "
                                      "between two different machines for "
                                      "calculation {}".format(calc.pk))

            remotedata = RemoteData(computer=computer,
                                    remote_path=workdir)
            remotedata.add_link_from(calc, label='remote_folder',
                                     link_type=LinkType.CREATE)
            remotedata.store()

            job_id = s.submit_from_script(t.getcwd(), script_filename)
            calc._set_job_id(job_id)
            # This should always be possible, because we should be
            # the only ones submitting this calculations,
            # so I do not check the ModificationNotAllowed
            calc._set_state(calc_states.WITHSCHEDULER)
            ## I do not set the state to queued; in this way, if the
            ## daemon is down, the user sees '(unknown)' as last state
            ## and understands that the daemon is not running.
            # if job_tmpl.submit_as_hold:
            #    calc._set_scheduler_state(job_states.QUEUED_HELD)
            #else:
            #    calc._set_scheduler_state(job_states.QUEUED)

            execlogger.debug("submitted calculation {} on {} with "
                             "jobid {}".format(calc.pk, computer.name, job_id),
                             extra=logger_extra)

    except Exception as e:
        import traceback

        try:
            calc._set_state(calc_states.SUBMISSIONFAILED)
        except ModificationNotAllowed:
            # Someone already set it, just skip
            pass

        execlogger.error("Submission of calc {} failed, check also the "
                         "log file! Traceback: {}".format(calc.pk,
                                                          traceback.format_exc()),
                         extra=logger_extra)
        raise
    finally:
        # close the transport, but only if it was opened within this function
        if must_open_t:
            t.close()


def retrieve_computed_for_authinfo(authinfo):
    from aiida.orm import JobCalculation
    from aiida.common.folders import SandboxFolder
    from aiida.orm.data.folder import FolderData
    from aiida.utils.logger import get_dblogger_extra
    from aiida.orm import DataFactory

    import os

    if not authinfo.enabled:
        return

    calcs_to_retrieve = list(JobCalculation._get_all_with_state(
        state=calc_states.COMPUTED,
        computer=authinfo.dbcomputer,
        user=authinfo.aiidauser)
    )
    retrieved = []

    # I avoid to open an ssh connection if there are no
    # calcs with state not COMPUTED
    if len(calcs_to_retrieve):

        # Open connection
        with authinfo.get_transport() as t:
            for calc in calcs_to_retrieve:
                logger_extra = get_dblogger_extra(calc)
                t._set_logger_extra(logger_extra)

                try:
                    calc._set_state(calc_states.RETRIEVING)
                except ModificationNotAllowed:
                    # Someone else has already started to retrieve it,
                    # just log and continue
                    execlogger.debug("Attempting to retrieve more than once "
                                     "calculation {}: skipping!".format(calc.pk),
                                     extra=logger_extra)
                    continue  # with the next calculation to retrieve
                try:
                    execlogger.debug("Retrieving calc {}".format(calc.pk),
                                     extra=logger_extra)
                    workdir = calc._get_remote_workdir()
                    retrieve_list = calc._get_retrieve_list()
                    retrieve_singlefile_list = calc._get_retrieve_singlefile_list()
                    execlogger.debug("[retrieval of calc {}] "
                                     "chdir {}".format(calc.pk, workdir),
                                     extra=logger_extra)
                    t.chdir(workdir)

                    retrieved_files = FolderData()
                    retrieved_files.add_link_from(
                        calc, label=calc._get_linkname_retrieved(),
                        link_type=LinkType.CREATE)

                    # First, retrieve the files of folderdata
                    with SandboxFolder() as folder:
                        for item in retrieve_list:
                            # I have two possibilities:
                            # * item is a string
                            # * or is a list
                            # then I have other two possibilities:
                            # * there are file patterns
                            # * or not
                            # First decide the name of the files
                            if isinstance(item, list):
                                tmp_rname, tmp_lname, depth = item
                                # if there are more than one file I do something differently
                                if t.has_magic(tmp_rname):
                                    remote_names = t.glob(tmp_rname)
                                    local_names = []
                                    for rem in remote_names:
                                        to_append = rem.split(os.path.sep)[-depth:] if depth > 0 else []
                                        local_names.append(os.path.sep.join([tmp_lname] + to_append))
                                else:
                                    remote_names = [tmp_rname]
                                    to_append = remote_names.split(os.path.sep)[-depth:] if depth > 0 else []
                                    local_names = [os.path.sep.join([tmp_lname] + to_append)]
                                if depth > 1:  # create directories in the folder, if needed
                                    for this_local_file in local_names:
                                        new_folder = os.path.join(
                                            folder.abspath,
                                            os.path.split(this_local_file)[0])
                                        if not os.path.exists(new_folder):
                                            os.makedirs(new_folder)
                            else:  # it is a string
                                if t.has_magic(item):
                                    remote_names = t.glob(item)
                                    local_names = [os.path.split(rem)[1] for rem in remote_names]
                                else:
                                    remote_names = [item]
                                    local_names = [os.path.split(item)[1]]

                            for rem, loc in zip(remote_names, local_names):
                                execlogger.debug("[retrieval of calc {}] "
                                                 "Trying to retrieve remote item '{}'".format(
                                    calc.pk, rem),
                                                 extra=logger_extra)
                                t.get(rem,
                                      os.path.join(folder.abspath, loc),
                                      ignore_nonexisting=True)

                        # Here I retrieved everything;
                        # now I store them inside the calculation
                        retrieved_files.replace_with_folder(folder.abspath,
                                                            overwrite=True)

                    # Second, retrieve the singlefiles
                    with SandboxFolder() as folder:
                        singlefile_list = []
                        for (linkname, subclassname, filename) in retrieve_singlefile_list:
                            execlogger.debug("[retrieval of calc {}] Trying "
                                             "to retrieve remote singlefile '{}'".format(
                                calc.pk, filename),
                                             extra=logger_extra)
                            localfilename = os.path.join(
                                folder.abspath, os.path.split(filename)[1])
                            t.get(filename, localfilename,
                                  ignore_nonexisting=True)
                            singlefile_list.append((linkname, subclassname,
                                                    localfilename))

                        # ignore files that have not been retrieved
                        singlefile_list = [i for i in singlefile_list if
                                           os.path.exists(i[2])]

                        # after retrieving from the cluster, I create the objects
                        singlefiles = []
                        for (linkname, subclassname, filename) in singlefile_list:
                            SinglefileSubclass = DataFactory(subclassname)
                            singlefile = SinglefileSubclass()
                            singlefile.set_file(filename)
                            singlefile.add_link_from(calc, label=linkname,
                                                     link_type=LinkType.CREATE)
                            singlefiles.append(singlefile)

                    # Finally, store
                    execlogger.debug("[retrieval of calc {}] "
                                     "Storing retrieved_files={}".format(
                        calc.pk, retrieved_files.dbnode.pk),
                                     extra=logger_extra)
                    retrieved_files.store()
                    for fil in singlefiles:
                        execlogger.debug("[retrieval of calc {}] "
                                         "Storing retrieved_singlefile={}".format(
                            calc.pk, fil.dbnode.pk),
                                         extra=logger_extra)
                        fil.store()

                    # If I was the one retrieving, I should also be the only
                    # one parsing! I do not check
                    calc._set_state(calc_states.PARSING)

                    Parser = calc.get_parserclass()
                    # If no parser is set, the calculation is successful
                    successful = True
                    if Parser is not None:
                        # TODO: parse here
                        parser = Parser(calc)
                        successful, new_nodes_tuple = parser.parse_from_calc()

                        for label, n in new_nodes_tuple:
                            n.add_link_from(calc, label=label,
                                            link_type=LinkType.CREATE)
                            n.store()

                    if successful:
                        try:
                            calc._set_state(calc_states.FINISHED)
                        except ModificationNotAllowed:
                            # I should have been the only one to set it, but
                            # in order to avoid unuseful error messages, I
                            # just ignore
                            pass
                    else:
                        try:
                            calc._set_state(calc_states.FAILED)
                        except ModificationNotAllowed:
                            # I should have been the only one to set it, but
                            # in order to avoid unuseful error messages, I
                            # just ignore
                            pass
                        execlogger.error("[parsing of calc {}] "
                                         "The parser returned an error, but it should have "
                                         "created an output node with some partial results "
                                         "and warnings. Check there for more information on "
                                         "the problem".format(calc.pk), extra=logger_extra)
                    retrieved.append(calc)
                except Exception:
                    import traceback

                    tb = traceback.format_exc()
                    newextradict = logger_extra.copy()
                    newextradict['full_traceback'] = tb
                    if calc.get_state() == calc_states.PARSING:
                        execlogger.error("Error parsing calc {}. "
                                         "Traceback: {}".format(calc.pk, tb),
                                         extra=newextradict)
                        # TODO: add a 'comment' to the calculation
                        try:
                            calc._set_state(calc_states.PARSINGFAILED)
                        except ModificationNotAllowed:
                            pass
                    else:
                        execlogger.error("Error retrieving calc {}. "
                                         "Traceback: {}".format(calc.pk, tb),
                                         extra=newextradict)
                        try:
                            calc._set_state(calc_states.RETRIEVALFAILED)
                        except ModificationNotAllowed:
                            pass
                        raise

    return retrieved



# -*- coding: utf-8 -*-
"""
This file provides very simple workflows for testing purposes.
Do not delete, otherwise 'verdi developertest' will stop to work.
"""
from aiida.orm.workflow import Workflow

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class WorkflowTestEmpty(Workflow):
    """
    Empty workflow, just for testing
    """

    def __init__(self, **kwargs):
        super(WorkflowTestEmpty, self).__init__(**kwargs)



# -*- coding: utf-8 -*-
"""
Tests for calculation nodes, attributes and links
"""

from aiida.backends.djsite.db.testbase import AiidaTestCase
from aiida.common.exceptions import ModificationNotAllowed


__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

class TestCalcNode(AiidaTestCase):
    """
    These tests check the features of Calculation nodes that differ from the
    base Node type
    """
    boolval = True
    intval = 123
    floatval = 4.56
    stringval = "aaaa"
    # A recursive dictionary
    dictval = {'num': 3, 'something': 'else', 'emptydict': {},
               'recursive': {'a': 1, 'b': True, 'c': 1.2, 'd': [1, 2, None],
                             'e': {'z': 'z', 'x': None, 'xx': {}, 'yy': []}}}
    listval = [1, "s", True, None]
    emptydict = {}
    emptylist = []

    def test_updatable_not_copied(self):
        """
        Checks the versioning.
        """
        from aiida.orm.test import myNodeWithFields

        # Has 'state' as updatable attribute
        a = myNodeWithFields()
        a._set_attr('state', 267)
        a.store()
        b = a.copy()

        # updatable attributes are not copied
        with self.assertRaises(AttributeError):
            b.get_attr('state')

    def test_delete_updatable_attributes(self):
        """
        Checks the versioning.
        """
        from aiida.orm.test import myNodeWithFields

        # Has 'state' as updatable attribute
        a = myNodeWithFields()
        attrs_to_set = {
            'bool': self.boolval,
            'integer': self.intval,
            'float': self.floatval,
            'string': self.stringval,
            'dict': self.dictval,
            'list': self.listval,
            'state': 267,  # updatable
        }

        for k, v in attrs_to_set.iteritems():
            a._set_attr(k, v)

        # Check before storing
        self.assertEquals(267, a.get_attr('state'))

        a.store()

        # Check after storing
        self.assertEquals(267, a.get_attr('state'))

        # Even if I stored many attributes, this should stay at 1
        self.assertEquals(a.dbnode.nodeversion, 1)

        # I should be able to delete the attribute
        a._del_attr('state')

        # I check increment on new version
        self.assertEquals(a.dbnode.nodeversion, 2)

        with self.assertRaises(AttributeError):
            # I check that I cannot modify this attribute
            _ = a.get_attr('state')

    def test_versioning_and_updatable_attributes(self):
        """
        Checks the versioning.
        """
        from aiida.orm.test import myNodeWithFields

        # Has 'state' as updatable attribute
        a = myNodeWithFields()
        attrs_to_set = {
            'bool': self.boolval,
            'integer': self.intval,
            'float': self.floatval,
            'string': self.stringval,
            'dict': self.dictval,
            'list': self.listval,
            'state': 267,
        }

        expected_version = 1

        for k, v in attrs_to_set.iteritems():
            a._set_attr(k, v)

        # Check before storing
        self.assertEquals(267, a.get_attr('state'))

        a.store()

        # Even if I stored many attributes, this should stay at 1
        self.assertEquals(a.dbnode.nodeversion, expected_version)

        # Sealing ups the version number
        a.seal()
        expected_version += 1
        self.assertEquals(a.dbnode.nodeversion, expected_version)

        # Check after storing
        self.assertEquals(267, a.get_attr('state'))
        self.assertEquals(a.dbnode.nodeversion, expected_version)

        # I check increment on new version
        a.set_extra('a', 'b')
        expected_version += 1
        self.assertEquals(a.dbnode.nodeversion, expected_version)

        # I check that I can set this attribute
        a._set_attr('state', 999)
        expected_version += 1

        # I check increment on new version
        self.assertEquals(a.dbnode.nodeversion, expected_version)

        with self.assertRaises(ModificationNotAllowed):
            # I check that I cannot modify this attribute
            a._set_attr('otherattribute', 222)

        # I check that the counter was not incremented
        self.assertEquals(a.dbnode.nodeversion, expected_version)

        # In both cases, the node version must increase
        a.label = 'test'
        expected_version += 1
        self.assertEquals(a.dbnode.nodeversion, expected_version)

        a.description = 'test description'
        expected_version += 1
        self.assertEquals(a.dbnode.nodeversion, expected_version)

        b = a.copy()
        # updatable attributes are not copied
        with self.assertRaises(AttributeError):
            b.get_attr('state')



# -*- coding: utf-8 -*-
from aiida.parsers.plugins.nwchem import BasenwcParser
from aiida.orm.calculation.job.nwchem.basic import BasicCalculation
from aiida.orm.data.parameter import ParameterData

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

class BasicParser(BasenwcParser):
    """
    Parser for the output of nwchem.
    """
    def __init__(self,calc):
        """
        Initialize the instance of BasicParser
        """
        # check for valid input
        self._check_calc_compatibility(calc)
        super(BasicParser, self).__init__(calc)

    def _check_calc_compatibility(self,calc):
        from aiida.common.exceptions import ParsingError
        if not isinstance(calc,BasicCalculation):
            raise ParsingError("Input calc must be a BasicCalculation")

    def _get_output_nodes(self, output_path, error_path):
        """
        Extracts output nodes from the standard output and standard error
        files.
        """
        from aiida.orm.data.array.trajectory import TrajectoryData
        import re

        state = None
        step = None
        scale = None
        with open(output_path) as f:
            lines = [x.strip('\n') for x in f.readlines()]

        result_dict = dict()
        trajectory = None
        for line in lines:
            if state is None and re.match('^\s*NWChem SCF Module\s*$',line):
                state = 'nwchem-scf-module'
                continue
            if state is None and re.match('^\s*NWChem Geometry Optimization\s*$',line):
                state = 'nwchem-geometry-optimisation'
                trajectory = TrajectoryData()
                continue
            if state == 'nwchem-scf-module' and re.match('^\s*Final RHF \s*results\s*$',line):
                state = 'final-rhf-results'
                continue
            if re.match('^\s*\-*\s*$',line):
                continue
            if state == 'final-rhf-results':
                result = re.match('^\s*([^=]+?)\s*=\s*([\-\d\.]+)$',line)
                if result:
                    key = re.sub('[^a-zA-Z0-9]+', '_', result.group(1).lower())
                    result_dict[key] = result.group(2)
                else:
                    state = 'nwchem-scf-module'
            if state == 'nwchem-geometry-optimisation' and re.match('^\s*Step\s+\d+\s*$',line):
                result = re.match('^\s*Step\s+(\d+)\s*$',line)
                step = result.group(1)
                continue
            if state == 'nwchem-geometry-optimisation' and \
                re.match('^\s*Output coordinates in a.u.',line):
                state = 'nwchem-geometry-optimisation-coordinates'
                result = re.match('scale by \s(*[\-\d\.]+)',line)
                scale = result.group(1)
                continue
        return [('parameters', ParameterData(dict=result_dict))]



# -*- coding: utf-8 -*-
import unittest

from aiida.backends.utils import load_dbenv, is_dbenv_loaded

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__authors__ = "The AiiDA team."
__version__ = "0.7.0.1"

if not is_dbenv_loaded():
    load_dbenv()

from aiida.workflows2.wf import wf
from aiida.workflows2.run import async
from aiida.workflows2.db_types import to_db_type


_True = to_db_type(True)


@wf
def simple_wf():
    return {'result': _True}

@wf
def return_input(inp):
    return {'result': inp}


class TestWf(unittest.TestCase):
    def test_blocking(self):
        self.assertTrue(simple_wf()['result'].value)
        self.assertTrue(return_input(_True)['result'].value)

    def test_async(self):
        self.assertTrue(async(simple_wf).result()['result'].value)
        self.assertTrue(async(return_input, _True).result()['result'].value)



# -*- coding: utf-8 -*-
"""
This is a simple plugin that takes two node inputs, both of type ParameterData,
with the following labels: template and parameters.
You can also add other SinglefileData nodes as input, that will be copied according to
what is written in 'template' (see below).

* parameters: a set of parameters that will be used for substitution.

* template: can contain the following parameters:

    * input_file_template: a string with substitutions to be managed with the format()\
      function of python, i.e. if you want to substitute a variable called 'varname', you write\
      {varname} in the text. See http://www.python.org/dev/peps/pep-3101/ for more\
      details. The replaced file will be the input file.

    * input_file_name: a string with the file name for the input. If it is not provided, no\
      file will be created.

    * output_file_name: a string with the file name for the output. If it is not provided, no\
      redirection will be done and the output will go in the scheduler output file.

    * cmdline_params: a list of strings, to be passed as command line parameters.\
      Each one is substituted with the same rule of input_file_template. Optional

    * input_through_stdin: if True, the input file name is passed via stdin. Default is\
      False if missing.

    * files_to_copy: if defined, a list of tuple pairs, with format ('link_name', 'dest_rel_path');\
         for each tuple, an input link to this calculation is looked for, with link labeled 'link_label',\
         and with file type 'Singlefile', and the content is copied to a remote file named 'dest_rel_path'\
         Errors are raised in the input links are non-existent, or of the wrong type, or if there are \
         unused input files.

TODO: probably use Python's Template strings instead??
TODO: catch exceptions
"""
from aiida.orm.calculation.job import JobCalculation
from aiida.common.exceptions import InputValidationError
from aiida.common.datastructures import CalcInfo, CodeInfo

# TODO: write a 'input_type_checker' routine to automatically check the existence
# and type of inputs + default values etc.

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class TemplatereplacerCalculation(JobCalculation):
    """
    Simple stub of a plugin that can be used to replace some text in a given
    template. Can be used for many different codes, or as a starting point
    to develop a new plugin.
    """

    def _prepare_for_submission(self, tempfolder, inputdict):
        """
        This is the routine to be called when you want to create
        the input files and related stuff with a plugin.

        :param tempfolder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        :param inputdict: a dictionary with the input nodes, as they would
                be returned by get_inputs_dict (with the Code!)
        """
        import StringIO

        from aiida.orm.data.parameter import ParameterData
        from aiida.orm.data.singlefile import SinglefileData
        from aiida.orm.data.remote import RemoteData
        from aiida.common.utils import validate_list_of_string_tuples
        from aiida.common.exceptions import ValidationError

        parameters_node = inputdict.pop('parameters', None)
        if parameters_node is None:
            parameters = {}
        else:
            if not isinstance(parameters_node, ParameterData):
                raise InputValidationError("'parameters' data is not of type ParameterData")
            parameters = dict(parameters_node.iterattrs())

        template_node = inputdict.pop('template', None)
        if template_node is None:
            raise InputValidationError("No 'template' input data")
        if not isinstance(template_node, ParameterData):
            raise InputValidationError("'template' data is not of type ParameterData")
        template = dict(template_node.iterattrs())

        input_file_template = template.pop('input_file_template', "")
        input_file_name = template.pop('input_file_name', None)
        output_file_name = template.pop('output_file_name', None)
        cmdline_params_tmpl = template.pop('cmdline_params', [])
        input_through_stdin = template.pop('input_through_stdin', False)
        files_to_copy = template.pop('files_to_copy', [])

        if template:
            raise InputValidationError("The following keys could not be "
                                       "used in the template node: {}".format(
                template.keys()))

        try:
            validate_list_of_string_tuples(files_to_copy, tuple_length=2)
        except ValidationError as e:
            raise InputValidationError("invalid file_to_copy format: {}".format(e.message))

        local_copy_list = []
        remote_copy_list = []

        for link_name, dest_rel_path in files_to_copy:
            try:
                fileobj = inputdict.pop(link_name)
            except KeyError:
                raise InputValidationError("You are asking to copy a file link {}, "
                                           "but there is no input link with such a name".format(link_name))
            if isinstance(fileobj, SinglefileData):
                local_copy_list.append((fileobj.get_file_abs_path(), dest_rel_path))
            elif isinstance(fileobj, RemoteData):  # can be a folder
                remote_copy_list.append(
                    (fileobj.get_computer().uuid, fileobj.get_remote_path(), dest_rel_path)
                )
            else:
                raise InputValidationError("If you ask to copy a file link {}, "
                                           "it must be either a SinglefileData or a RemoteData; it is instead of type {}".format(
                    link_name, fileobj.__class__.__name__))

        code = inputdict.pop('code', None)
        if code is None:
            raise InputValidationError("No code in input")

        if len(inputdict) > 0:
            raise InputValidationError("The input nodes with the following labels could not be "
                                       "used by the templatereplacer plugin: {}".format(
                inputdict.keys()))

        if input_file_name is not None and not input_file_template:
            raise InputValidationError("If you give an input_file_name, you "
                                       "must also specify a input_file_template")

        if input_through_stdin and input_file_name is None:
            raise InputValidationError("If you ask for input_through_stdin you have to "
                                       "specify a input_file_name")

        input_file = StringIO.StringIO(input_file_template.format(**parameters))
        if input_file_name:
            tempfolder.create_file_from_filelike(input_file, input_file_name)
        else:
            if input_file_template:
                self.logger.warning("No input file name passed, but a input file template is present")

        cmdline_params = [i.format(**parameters) for i in cmdline_params_tmpl]

        calcinfo = CalcInfo()
        calcinfo.retrieve_list = []

        calcinfo.uuid = self.uuid
        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = remote_copy_list

        codeinfo = CodeInfo()
        codeinfo.cmdline_params = cmdline_params
        if input_through_stdin is not None:
            codeinfo.stdin_name = input_file_name
        if output_file_name:
            codeinfo.stdout_name = output_file_name
            calcinfo.retrieve_list.append(output_file_name)
        codeinfo.code_uuid = code.uuid
        calcinfo.codes_info = [codeinfo]

        return calcinfo



# -*- coding: utf-8 -*-

import importlib
from threading import local

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class ProcessStack(object):
    # Use thread-local storage for the stack
    _thread_local = local()

    @classmethod
    def top(cls):
        return cls.stack()[-1]

    @classmethod
    def stack(cls):
        try:
            return cls._thread_local.wf_stack
        except AttributeError:
            cls._thread_local.wf_stack = []
            return cls._thread_local.wf_stack

    @classmethod
    def push(cls, process):
        try:
            process._parent = cls.top()
        except IndexError:
            process._parent = None
        cls.stack().append(process)

    @classmethod
    def pop(cls):
        process = cls.stack().pop()
        process._parent = None

    def __init__(self):
        raise NotImplementedError("Can't instantiate the ProcessStack")


def load_class(classstring):
    """
    Load a class from a string
    """
    class_data = classstring.split(".")
    module_path = ".".join(class_data[:-1])
    class_name = class_data[-1]

    module = importlib.import_module(module_path)
    # Finally, we retrieve the Class
    return getattr(module, class_name)


def is_workfunction(func):
    try:
        return func._is_workfunction
    except AttributeError:
        return False



# -*- coding: utf-8 -*-

from aiida.orm.calculation.job import JobCalculation
from aiida.orm.data.cif import CifData
from aiida.orm.data.parameter import ParameterData
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.common.exceptions import InputValidationError
from aiida.common.utils import classproperty

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."


class CiffilterCalculation(JobCalculation):
    """
    Generic input plugin for scripts from cod-tools package.
    """

    def _init_internal_params(self):
        super(CiffilterCalculation, self)._init_internal_params()

        # Name of the default parser
        self._default_parser = 'codtools.ciffilter'

        # Default command line parameters
        self._default_commandline_params = []

        # Default input and output files
        self._DEFAULT_INPUT_FILE = 'aiida.in'
        self._DEFAULT_OUTPUT_FILE = 'aiida.out'
        self._DEFAULT_ERROR_FILE = 'aiida.err'

    def set_resources(self, resources_dict):
        """
        Overrides the original ``set_resouces()`` in order to prevent
        parallelization, which is not supported and may cause strange
        behaviour.

        :raises FeatureNotAvailable: when ``num_machines`` or
            ``num_mpiprocs_per_machine`` are being set to something other
            than 1.
        """
        self._validate_resources(**resources_dict)
        super(CiffilterCalculation, self).set_resources(resources_dict)

    @classproperty
    def _use_methods(cls):
        retdict = JobCalculation._use_methods
        retdict.update({
            "cif": {
                'valid_types': CifData,
                'additional_parameter': None,
                'linkname': 'cif',
                'docstring': "A CIF file to be processed",
            },
            "parameters": {
                'valid_types': ParameterData,
                'additional_parameter': None,
                'linkname': 'parameters',
                'docstring': "Parameters used in command line",
            },
        })
        return retdict

    def _validate_resources(self, **kwargs):
        from aiida.common.exceptions import FeatureNotAvailable

        for key in ['num_machines', 'num_mpiprocs_per_machine',
                    'tot_num_mpiprocs']:
            if key in kwargs and kwargs[key] != 1:
                raise FeatureNotAvailable(
                    "Cannot set resouce '{}' to value '{}' for {}: "
                    "parallelization is not supported, only value of "
                    "'1' is accepted.".format(key, kwargs[key],
                                              self.__class__.__name__))

    def _prepare_for_submission(self, tempfolder, inputdict):
        from aiida.orm.calculation.job.codtools import commandline_params_from_dict
        import shutil

        try:
            cif = inputdict.pop(self.get_linkname('cif'))
        except KeyError:
            raise InputValidationError("no CIF file is specified for this calculation")
        if not isinstance(cif, CifData):
            raise InputValidationError("cif is not of type CifData")

        parameters = inputdict.pop(self.get_linkname('parameters'), None)
        if parameters is None:
            parameters = ParameterData(dict={})
        if not isinstance(parameters, ParameterData):
            raise InputValidationError("parameters is not of type ParameterData")

        code = inputdict.pop(self.get_linkname('code'), None)
        if code is None:
            raise InputValidationError("Code not found in input")

        self._validate_resources(**self.get_resources())

        input_filename = tempfolder.get_abs_path(self._DEFAULT_INPUT_FILE)
        shutil.copy(cif.get_file_abs_path(), input_filename)

        commandline_params = self._default_commandline_params
        commandline_params.extend(
            commandline_params_from_dict(parameters.get_dict()))

        calcinfo = CalcInfo()
        calcinfo.uuid = self.uuid
        # The command line parameters should be generated from 'parameters'
        calcinfo.local_copy_list = []
        calcinfo.remote_copy_list = []
        calcinfo.retrieve_list = [self._DEFAULT_OUTPUT_FILE,
                                  self._DEFAULT_ERROR_FILE]
        calcinfo.retrieve_singlefile_list = []

        codeinfo = CodeInfo()
        codeinfo.cmdline_params = commandline_params
        codeinfo.stdin_name = self._DEFAULT_INPUT_FILE
        codeinfo.stdout_name = self._DEFAULT_OUTPUT_FILE
        codeinfo.stderr_name = self._DEFAULT_ERROR_FILE
        codeinfo.code_uuid = code.uuid
        calcinfo.codes_info = [codeinfo]
        
        return calcinfo
